import jax
import jax.numpy as jnp
import chex
from flax.struct import dataclass
from typing import Tuple
from jaxmarl import make
from jaxmarl.gridworld.ma_maze import MAMaze
import math
from functools import partial
from jaxmarl.gridworld.grid_viz import GridVisualizer

@dataclass
class ArrayFixedSizeSet:
    """Fairly ridiculous class that can map from discrete arrays to unique integers.
    It needs linear time to both insert and retrieve elements, but does guarantee
    static sizing and can be passed to jit-ed functions etc. Uses a fixed placeholder
    value that must never occur.
    """

    data: chex.Array
    capacity: int
    array_shape: Tuple[int]
    size: int
    placeholder: int = 0

    @classmethod
    def create(cls, array_shape: Tuple[int], capacity: int, placeholder: int = 0):
        data = jnp.zeros((capacity, *array_shape)) + placeholder
        return cls(
            data=data,
            capacity=capacity,
            array_shape=array_shape,
            size=0,
            placeholder=placeholder,
        )

    def put(self, arr: chex.Array):
        idx = self.get(arr)
        data = jax.lax.cond(
            idx == -1, lambda: self.data.at[self.size].set(arr), lambda: self.data
        )
        size = jax.lax.cond(
            idx == -1,
            lambda: self.size + 1,
            lambda: self.size,
        )
        return self.replace(data=data, size=size)

    def get(self, arr: chex.Array):
        idx = jnp.argwhere(
            jnp.all(self.data == arr, axis=range(1, self.data.ndim)),
            size=1,
            fill_value=-1,
        )
        return idx.squeeze()


def main():

    epsilon = 0.2
    alpha = 0.1
    n_actions = 5
    gamma = 0.99
    env = MAMaze(height=3, width=3, n_walls=2)
    env_params = env.default_params
    buffer_capacity = 10000
    n_agents = 2
    rng = jax.random.PRNGKey(0)
    total_timesteps = 50000
    eval_timesteps = 1000

    def flatten_obs(obs):
        img = obs['image'].reshape((*obs['image'].shape[:-3], -1))
        return jnp.concatenate((img, jnp.expand_dims(obs['agent_dir'], 0)), axis=-1)

    def extract_obs(obs, idx):
        return jax.tree.map(lambda x: x[idx], obs)

    def sample_random_action(key):
        return jax.random.randint(key, shape=(), minval=0, maxval=n_actions)

    def sample_optimal_action(key, obs_storage, q_values, obs):
        q_index = obs_storage.get(flatten_obs(extract_obs(obs, 0)))
        max_action = jnp.argmax(q_values[q_index]).squeeze()
        random_action = sample_random_action(key)
        return jax.lax.cond(jnp.all(q_values[q_index] == q_values[q_index][0]), lambda: random_action, lambda: max_action)
    
    def sample_action(key, obs_storage, q_values, obs):

        key, key_rand_action, key_optimal_action = jax.random.split(key, num=3)
        optimal_action = sample_optimal_action(key=key_optimal_action, obs_storage=obs_storage, q_values=q_values, obs=obs)
        random_action = sample_random_action(key_rand_action)
        return jax.lax.cond(
            jax.random.uniform(key) < epsilon, lambda: random_action, lambda: optimal_action
        )

    @partial(jax.jit, static_argnums=(2,))
    def step(carry, _, evaluate=False):
        obs_storage, q_values, state, obs, cum_reward, key = carry
        key, action_key, opt_action_key = jax.random.split(key, num=3)
        action_key = jax.random.split(action_key, num=n_agents)
        if not evaluate:
            sampled_action = sample_action(opt_action_key, obs_storage, q_values, obs)
        else:
            sampled_action = sample_optimal_action(opt_action_key, obs_storage, q_values, obs)
        # actions = {
        #     agent: sample_random_action(action_key[i])
        #     for i, agent in enumerate(env.agents)
        # }
        # actions[env.agents[0]] = sampled_action
        actions = jnp.array([sampled_action, sample_random_action(action_key[1])])
        key, state_key = jax.random.split(key)
        next_obs, state, rewards, dones, infos = env.step(state_key, state, actions)
        cum_reward += rewards

        if not evaluate:
            obs_storage = obs_storage.put(flatten_obs(extract_obs(next_obs, 0)))
            q_index = obs_storage.get(flatten_obs(extract_obs(obs, 0)))
            next_q_index = obs_storage.get(flatten_obs(extract_obs(next_obs, 0)))
            updated_q_values = q_values.at[q_index, sampled_action].set(
                q_values[q_index, sampled_action]
                + alpha
                * (
                    gamma * jnp.max(q_values[next_q_index])
                    + rewards
                    - q_values[q_index, sampled_action]
                )
            )
            q_values = jax.lax.cond(dones, lambda: q_values, lambda: updated_q_values)
        return (obs_storage, q_values, state, obs, cum_reward, key), _


    obs_shape = math.prod(env.obs_shape) + 1
    obs_storage = ArrayFixedSizeSet.create((obs_shape,), capacity=buffer_capacity, placeholder=-1)
    q_values = jnp.zeros(shape=(buffer_capacity, n_actions))
    rng, rng_reset = jax.random.split(rng)
    obs, state = env.reset(rng_reset)
    init = (obs_storage, q_values, state, obs, 0, rng)
    (obs_storage, q_values, state, obs, cum_reward, rng), _ = jax.lax.scan(partial(step, evaluate=False), init=init, xs=None, length=total_timesteps)
    print(f"Total reward: {cum_reward}")
    rng, eval_rng_reset = jax.random.split(rng)
    obs, state = env.reset(eval_rng_reset)
    eval_init = (obs_storage, q_values, state, obs, 0, rng)
    (obs_storage, q_values, state, obs, cum_reward, rng), _ = jax.lax.scan(partial(step, evaluate=True), init=eval_init, xs=None, length=eval_timesteps)
    print(f"Total reward: {cum_reward}")
    viz = GridVisualizer()
    obs, state = env.reset(eval_rng_reset)
    viz.render(env_params, state, highlight=False)
    # TODO visualise the policy.
    for _ in range(1000):
        rng, action_key, opt_action_key = jax.random.split(rng, num=3)
        sampled_optimal_action = sample_optimal_action(opt_action_key, obs_storage, q_values, obs)
        actions = jnp.array([sampled_optimal_action, sample_random_action(action_key)])
        rng, state_key = jax.random.split(rng)
        obs, state, rewards, dones, infos = env.step(state_key, state, actions)
        viz.render(env_params, state, highlight=False)


if __name__ == "__main__":
    main()



import jax
from jax import numpy as jnp
from jaxmarl import make
import random
import pprint
import sys
import numpy as np
import argparse
from jaxmarl.wrappers.baselines import load_params
from jaxmarl.environments.hanabi.pretrained.obl_r2d2_agent import OBLAgentR2D2

env = make("hanabi")
batchify = lambda x: jnp.stack([x[agent] for agent in env.agents])
unbatchify = lambda x: {agent: x[i] for i, agent in enumerate(env.agents)}


class ManualPlayer:
    def __init__(self, player_idx):
        self._player_idx = player_idx

    def act(self, obs, legal_moves, curr_player) -> int:
        legal_moves = batchify(legal_moves)
        actions = np.array([20, 20])

        print("valid actions:", [(a, env.action_encoding[a]) for a in np.where(legal_moves[curr_player] == 1)[0]])

        if curr_player != self._player_idx:
            return actions

        # take action input from user
        while True:
            try:
                print("---")
                action = int(input("Insert manual action: "))
                print("action legal:", legal_moves[curr_player][action])
                print("---\n")
                if (
                    action >= 0
                    and action <= 20
                    and legal_moves[curr_player][action] == 1
                ):
                    break
                else:
                    print("Invalid action.")
            except KeyboardInterrupt:
                sys.exit(0)
            except:
                action = 0
                print("Invalid action.")

        actions[curr_player] = action

        return actions


class ManualGameOBLAgentR2D2:
    def __init__(self, weight_file, player_idx):
        self.player_id = player_idx
        self.params = load_params(weight_file)
        self.model = OBLAgentR2D2()
        self.carry = self.model.initialize_carry(jax.random.PRNGKey(0), batch_dims=(2,))

    def act(self, obs, legal_moves, curr_player):
        obs = batchify(obs)
        legal_moves = batchify(legal_moves)
        self.carry, actions = self.model.greedy_act(
            self.params, self.carry, (obs, legal_moves)
        )
        return actions


def get_agents(args):
    agents = []
    for player_idx in [0, 1]:
        player_type = getattr(args, f"player{player_idx}")
        weight_file = getattr(args, f"weight{player_idx}")
        if player_type == "manual":
            agents.append(ManualPlayer(player_idx))
        elif player_type == "obl":
            assert (
                weight_file is not None
            ), "Weight file must be provided for all the OBL agents."
            agents.append(ManualGameOBLAgentR2D2(weight_file, player_idx))
    return agents


def play_game(args):
    if args.seed is not None:
        seed = args.seed
    else:
        seed = random.randint(0, 10000)
    print(f"{'-'*10}\nStarting new game with random seed: {seed}\n")

    agents = get_agents(args)

    use_jit = args.use_jit if args.use_jit is not None else True
    with jax.disable_jit(not use_jit):

        rng = jax.random.PRNGKey(seed)
        rng, _rng = jax.random.split(rng)

        obs, env_state = env.reset(_rng)
        legal_moves = env.get_legal_moves(env_state)

        @jax.jit
        def _step_env(rng, env_state, actions):
            rng, _rng = jax.random.split(rng)
            new_obs, new_env_state, reward, dones, infos = env.step(
                _rng, env_state, actions
            )
            new_legal_moves = env.get_legal_moves(new_env_state)
            return rng, new_env_state, new_obs, reward, dones, new_legal_moves

        done = False
        cum_rew = 0
        t = 0

        print("\n" + "=" * 40 + "\n")

        while not done:
            env.render(env_state)

            curr_player = np.where(env_state.cur_player_idx == 1)[0][0]
            actions_all = [
                agents[i].act(obs, legal_moves, curr_player)
                for i in range(len(env.agents))
            ]

            actions = actions_all[curr_player]
            print(f"Action played: {env.action_encoding[int(actions[curr_player])]}")

            actions = {
                agent: jnp.array(actions[i]) for i, agent in enumerate(env.agents)
            }

            rng, env_state, obs, reward, dones, legal_moves = _step_env(
                rng, env_state, actions
            )

            done = dones["__all__"]
            cum_rew += reward["__all__"]
            t += 1

            print("\n" + "=" * 40 + "\n")

        print("Game Ended. Score:", cum_rew)


def main(args):
    print("Starting Hanabi. Remember, actions encoding is:")
    pprint.pprint(env.action_encoding)
    play_game(args)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--player0", type=str, default="obl")
    parser.add_argument("--player1", type=str, default="obl")
    parser.add_argument("--weight0", type=str, default=None)
    parser.add_argument("--weight1", type=str, default=None)
    parser.add_argument("--seed", type=int, default=None)
    parser.add_argument("--use_jit", type=bool, default=True)
    args = parser.parse_args()
    print(args)
    return args


if __name__ == "__main__":
    args = parse_args()
    main(args)



import jax
import chex
from typing import Tuple, Union, Optional
from functools import partial
from flax import struct


@struct.dataclass
class EnvState:
	time: int


@struct.dataclass
class EnvParams:
	max_episode_steps: int


class Environment(object):
	@property
	def default_params(self) -> EnvParams:
		return EnvParams()

	@partial(jax.jit, static_argnums=(0,4))
	def step(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: Union[int, float]
	) -> Tuple[chex.ArrayTree, EnvState, float, bool]:
		"""Performs step transitions in the environment."""
		# Use default env parameters if no others specified
		if hasattr(self, 'params'):
			params = self.params
		else:
			params = self.default_params

		key, key_reset = jax.random.split(key)
		obs_st, state_st, reward, done, info = self.step_env(
			key, state, action
		)

		if hasattr(params, 'singleton_seed') \
			and params.singleton_seed >= 0:
			key_reset = jax.random.PRNGKey(params.singleton_seed)

		obs_re, state_re = self.reset_env(key_reset)

		# Auto-reset environment based on termination
		state = jax.tree.map(
			lambda x, y: jax.lax.select(done, x, y), state_re, state_st
		)
		obs = jax.tree.map(
			lambda x, y: jax.lax.select(done, x, y), obs_re, obs_st
		)

		return obs, state, reward, done, info

	@partial(jax.jit, static_argnums=(0,))
	def reset(
		self, 
		key: chex.PRNGKey, 
	) -> Tuple[chex.ArrayTree, EnvState]:
		"""Performs resetting of environment."""
		# Use default env parameters if no others specified
		if hasattr(self, 'params'):
			params = self.params
		else:
			params = self.default_params

		if hasattr(params, 'singleton_seed') \
			and params.singleton_seed >= 0:
			key = jax.random.PRNGKey(params.singleton_seed)
		obs, state = self.reset_env(key)
		return obs, state

	def step_env(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: Union[int, float],
	) -> Tuple[chex.ArrayTree, EnvState, float, bool, dict]:
		"""Environment-specific step transition."""
		raise NotImplementedError

	def reset_env(
		self, key: chex.PRNGKey
	) -> Tuple[chex.ArrayTree, EnvState]:
		"""Environment-specific reset."""
		raise NotImplementedError

	def get_obs(self, state: EnvState) -> chex.ArrayTree:
		"""Applies observation function to state."""
		raise NotImplementedError

	def is_terminal(self, state: EnvState) -> bool:
		"""Check whether state is terminal."""
		raise NotImplementedError

	@property
	def name(self) -> str:
		"""Environment name."""
		return type(self).__name__

	@property
	def num_actions(self) -> int:
		"""Number of actions possible in environment."""
		raise NotImplementedError

	def action_space(self):
		"""Action space of the environment."""
		raise NotImplementedError

	def observation_space(self):
		"""Observation space of the environment."""
		raise NotImplementedError

	def state_space(self):
		"""State space of the environment."""
		raise NotImplementedError

	def max_episode_steps(self):
		"""Maximum number of time steps in environment."""
		raise NotImplementedError



import argparse
import jax
import jax.numpy as jnp
from jaxmarl.environments.overcooked_v2.common import Actions
from jaxmarl.environments.overcooked_v2.overcooked import OvercookedV2
from jaxmarl.environments.overcooked_v2.layouts import overcooked_v2_layouts as layouts
from jaxmarl.viz.overcooked_v2_visualizer import OvercookedV2Visualizer


class InteractiveOvercookedV2:

    def __init__(self, layout, agent_view_size=None, no_jit=False, debug=False):
        self.debug = debug
        self.no_jit = no_jit

        self.env = OvercookedV2(layout=layout, agent_view_size=agent_view_size)
        self.viz = OvercookedV2Visualizer()

    def run(self, key):
        self.key = key
        with jax.disable_jit(self.no_jit):
            self._run()

    def _run(self):
        self._reset()

        self.viz.window.reg_key_handler(self._handle_input)
        self.viz.show(block=True)

    def _handle_input(self, event):
        if self.debug:
            print("Pressed", event.key)

        ACTION_MAPPING = {
            "left": Actions.left,
            "right": Actions.right,
            "up": Actions.up,
            "down": Actions.down,
            " ": Actions.interact,
            "tab": Actions.stay,
        }

        match event.key:
            case "escape":
                self.viz.window.close()
                return
            case "backspace":
                self._reset()
                return
            case key if key in ACTION_MAPPING:
                action = ACTION_MAPPING[key]
            case key:
                print(f"Key {key} not recognized")
                return

        self._step(action)

    def _redraw(self):
        self.viz.render(self.state, agent_view_size=self.env.agent_view_size)

    def _reset(self):
        self.key, key = jax.random.split(self.key)
        _, state = jax.jit(self.env.reset)(key)
        self.state = state

        self._redraw()

    def _step(self, action):
        self.key, subkey = jax.random.split(self.key)

        actions = {f"agent_{i}": jnp.array(action) for i in range(self.env.num_agents)}
        if self.debug:
            print("Actions: ", actions)

        obs, state, reward, done, info = jax.jit(self.env.step_env)(
            subkey, self.state, actions
        )
        self.state = state
        print(f"t={state.time}: reward={reward['agent_0']}, done = {done['__all__']}")

        if self.debug:
            a0_obs = obs["agent_0"]
            a0_obs = jnp.transpose(a0_obs, (2, 0, 1))
            print("Agent 0 observation: ", a0_obs)
            print("Reward: ", reward)
            print("Shaped reward: ", info["shaped_reward"])

        if done["__all__"]:
            self._reset()
        else:
            self._redraw()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--layout", type=str, help="Overcooked layout", default="cramped_room"
    )
    parser.add_argument(
        "--seed",
        type=int,
        help="random seed to generate the environment with",
        default=0,
    )
    parser.add_argument(
        "--agent_view_size",
        type=int,
        help="Number of cells in the agent view. If not provided, the agent will see the whole grid.",
    )
    parser.add_argument(
        "--no_jit",
        default=False,
        help="Disable JIT compilation",
        action="store_true",
    )
    parser.add_argument(
        "--debug", default=False, help="Debug mode", action="store_true"
    )
    args = parser.parse_args()

    if len(args.layout) == 0:
        raise ValueError("You must provide a layout.")
    layout = layouts[args.layout]

    interactive = InteractiveOvercookedV2(
        layout=layout,
        agent_view_size=args.agent_view_size,
        no_jit=args.no_jit,
        debug=args.debug,
    )

    key = jax.random.PRNGKey(args.seed)
    interactive.run(key)



import math
from enum import IntEnum
from typing import Any, Optional, Tuple, Union, List, Dict
import itertools

import chex
import jax
import jax.numpy as jnp
from jax import jit
from functools import partial
import numpy as onp

from jaxmarl.environments.multi_agent_env import MultiAgentEnv
from jaxmarl.environments import spaces
from flax.struct import dataclass

from .rendering import (
    downsample,
    fill_coords,
    highlight_img,
    point_in_circle,
    point_in_rect,
    point_in_triangle,
    rotate_fn,
)

GRID_SIZE = 8
OBS_SIZE = 5
PADDING = OBS_SIZE - 1
NUM_TYPES = 5  # empty (0), red (1), blue, red coin, blue coin, wall, interact
NUM_COINS = 6  # per type
NUM_COIN_TYPES = 2
NUM_OBJECTS = (
    2 + NUM_COIN_TYPES * NUM_COINS + 1
)  # red, blue, 2 red coin, 2 blue coin

INTERACT_THRESHOLD = 0


@dataclass
class State:
    # n players x 2
    agent_positions: List[jnp.ndarray]
    inner_t: int
    outer_t: int
    grid: jnp.ndarray
    # n players x 2
    agent_inventories: List[jnp.ndarray]
    # coin positions
    coin_coop: List[jnp.ndarray]
    coin_defect: List[jnp.ndarray]
    agent_freezes: List[int]
    coin_timer_coop: List[int]
    coin_timer_defect: List[int]


class Actions(IntEnum):
    left = 0
    right = 1
    forward = 2
    interact = 3
    stay = 4


class Items(IntEnum):
    empty = 0
    agent1 = 1
    agent2 = 2
    agent3 = 3
    agent4 = 4
    agent5 = 5
    agent6 = 6
    agent7 = 7
    agent8 = 8
    cooperation_coin = 9
    defection_coin = 10
    wall = 11
    interact = 12
    #agents if  1 < val < item.WALL 

AGENT_IDX = jnp.array([1,2,3,4,5,6,7,8])
ROTATIONS = jnp.array(
    [
        [0, 0, 1],  # turn left
        [0, 0, -1],  # turn right
        [0, 0, 0],  # forward
        [0, 0, 0],  # stay
        [0, 0, 0],  # zap`
    ],
    dtype=jnp.int8,
)

STEP = jnp.array(
    [
        [0, 1, 0],  # up
        [1, 0, 0],  # right
        [0, -1, 0],  # down
        [-1, 0, 0],  # left
    ],
    dtype=jnp.int8,
)

GRID = jnp.zeros(
    (GRID_SIZE + 2 * PADDING, GRID_SIZE + 2 * PADDING),
    dtype=jnp.int8,
)

# First layer of Padding is Wall
GRID = GRID.at[PADDING - 1, :].set(5)
GRID = GRID.at[GRID_SIZE + PADDING, :].set(5)
GRID = GRID.at[:, PADDING - 1].set(5)
GRID = GRID.at[:, GRID_SIZE + PADDING].set(5)

COIN_SPAWNS = [
    [1, 1],
    [1, 2],
    [2, 1],
    [1, GRID_SIZE - 2],
    [2, GRID_SIZE - 2],
    [1, GRID_SIZE - 3],
    # [2, 2],
    # [2, GRID_SIZE - 3],
    [GRID_SIZE - 2, 2],
    [GRID_SIZE - 3, 1],
    [GRID_SIZE - 2, 1],
    [GRID_SIZE - 2, GRID_SIZE - 2],
    [GRID_SIZE - 2, GRID_SIZE - 3],
    [GRID_SIZE - 3, GRID_SIZE - 2],
    # [GRID_SIZE - 3, 2],
    # [GRID_SIZE - 3, GRID_SIZE - 3],
]

COIN_SPAWNS = jnp.array(
    COIN_SPAWNS,
    dtype=jnp.int8,
)

RED_SPAWN = jnp.array(
    COIN_SPAWNS[::2, :],
    dtype=jnp.int8,
)

BLUE_SPAWN = jnp.array(
    COIN_SPAWNS[1::2, :],
    dtype=jnp.int8,
)

AGENT_SPAWNS = [
    [0, 0],
    [0, 1],
    [0, 2],
    [1, 0],
    # [1, 1],
    [1, 2],
    [0, GRID_SIZE - 1],
    [0, GRID_SIZE - 2],
    [0, GRID_SIZE - 3],
    [1, GRID_SIZE - 1],
    # [1, GRID_SIZE - 2],
    # [1, GRID_SIZE - 3],
    [GRID_SIZE - 1, 0],
    [GRID_SIZE - 1, 1],
    [GRID_SIZE - 1, 2],
    [GRID_SIZE - 2, 0],
    # [GRID_SIZE - 2, 1],
    # [GRID_SIZE - 2, 2],
    [GRID_SIZE - 1, GRID_SIZE - 1],
    [GRID_SIZE - 1, GRID_SIZE - 2],
    [GRID_SIZE - 1, GRID_SIZE - 3],
    [GRID_SIZE - 2, GRID_SIZE - 1],
    # [GRID_SIZE - 2, GRID_SIZE - 2],
    [GRID_SIZE - 2, GRID_SIZE - 3],
]

# Get all combinations of AGENT_SPAWNS with length 8
agent_spawn_combinations = list(itertools.combinations(AGENT_SPAWNS, 8))

# Convert the list of combinations to a JAX numpy array
AGENT_SPAWNS = jnp.array(agent_spawn_combinations, dtype=jnp.int8)


PLAYER1_COLOUR = (255.0, 127.0, 14.0) #kinda orange
PLAYER2_COLOUR = (31.0, 119.0, 180.0) #kinda blue
PLAYER3_COLOUR = (236.0, 64.0, 122.0) #kinda pink
PLAYER4_COLOUR = (255.0, 235.0, 59.0) #yellow
PLAYER5_COLOUR = (41.0, 182.0, 246.0) #baby blue
PLAYER6_COLOUR = (171.0, 71.0, 188.0) #purple
PLAYER7_COLOUR = (121.0, 85.0, 72.0) #brown
PLAYER8_COLOUR = (255.0, 205.0, 210.0) #salmon
GREEN_COLOUR = (44.0, 160.0, 44.0)
RED_COLOUR = (214.0, 39.0, 40.0)



class InTheGrid(MultiAgentEnv):
    """
    JAX Compatible version of *inTheMatix environment.
    """

    # used for caching
    tile_cache: Dict[Tuple[Any, ...], Any] = {}

    def __init__(
        self,
        num_inner_steps=152,
        num_outer_steps=1,
        num_agents=2,
        fixed_coin_location=True,
        payoff_matrix=jnp.array([[[3, 0], [5, 1]], [[3, 5], [0, 1]]]),
        freeze_penalty=5,
    ):

        super().__init__(num_agents=num_agents)
        self.agents = list(range(num_agents))

        def _get_obs_point(x: int, y: int, dir: int) -> jnp.ndarray:
            x, y = x + PADDING, y + PADDING
            x = jnp.where(dir == 0, x - (OBS_SIZE // 2), x)
            x = jnp.where(dir == 2, x - (OBS_SIZE // 2), x)
            x = jnp.where(dir == 3, x - (OBS_SIZE - 1), x)

            y = jnp.where(dir == 1, y - (OBS_SIZE // 2), y)
            y = jnp.where(dir == 2, y - (OBS_SIZE - 1), y)
            y = jnp.where(dir == 3, y - (OBS_SIZE // 2), y)
            return x, y

        def _get_obs(state: State) -> jnp.ndarray:
            # create state
            grid = jnp.pad(
                state.grid,
                ((PADDING, PADDING), (PADDING, PADDING)),
                constant_values=Items.wall,
            )
            def mini_obs(agent_pos, agent_idx, other_agents_idx, other_agents_pos):
                x, y = _get_obs_point(
                agent_pos[0], agent_pos[1], agent_pos[2]
                )
                grid1 = jax.lax.dynamic_slice(
                    grid,
                    start_indices=(x, y),
                    slice_sizes=(OBS_SIZE, OBS_SIZE),
                )
                # rotate
                grid1 = jnp.where(
                    agent_pos[2] == 1,
                    jnp.rot90(grid1, k=1, axes=(0, 1)),
                    grid1,
                )
                grid1 = jnp.where(
                    agent_pos[2] == 2,
                    jnp.rot90(grid1, k=2, axes=(0, 1)),
                    grid1,
                )
                grid1 = jnp.where(
                    agent_pos[2] == 3,
                    jnp.rot90(grid1, k=3, axes=(0, 1)),
                    grid1,
                )
                # angle1 = -1 * jnp.ones_like(grid1, dtype=jnp.int8)
            #      i look where the blue agent is (with its own orientation)
            #     i correct its orientation relative to mine
            #     angle is 

                def check_angle(grid1, item, other_agent_pos):    
                    angle1 = jnp.where(
                        grid1 == item,
                        ((other_agent_pos[2] - agent_pos[2])+1) % 5,
                        0,
                    )
                    return angle1
                
                vmap_check_angle = jax.vmap(check_angle, (None, 0, 0), 0)
                angle1 = vmap_check_angle(grid1, other_agents_idx, other_agents_pos)
                angle1 = jnp.sum(angle1, axis=0) - 1
                angle1 = jax.nn.one_hot(angle1, 4)
                # one-hot (drop first channel as its empty blocks)
                grid1 = jax.nn.one_hot(grid1 - 1, len(Items) - 1, dtype=jnp.int8)
                _grid1 = grid1.at[:, :, 0].set(grid1[:, :, agent_idx])
                _grid1 = _grid1.at[:, :, agent_idx].set(grid1[:, :, 0])
                obs1 = jnp.concatenate([_grid1, angle1], axis=-1)
                return obs1

            vmap_mini_obs = jax.vmap(mini_obs, (0, 0, 0, 0), 0)
            agent_indices = jnp.arange(1, num_agents+1)
            expanded_indices = agent_indices - 1
            expanded_indices = expanded_indices[:, jnp.newaxis]
            # Create a matrix with the desired shape and fill it with the expanded_indices
            other_agent_indices = expanded_indices + jnp.arange(1, num_agents)
            # Correct the indices by adding 1 where the agent's own index is excluded
            other_agent_indices = other_agent_indices + (other_agent_indices >= expanded_indices)

            # Create an array of agent indices
            agent_indices = jnp.arange(num_agents)

            # Add a new axis to agent_indices and agent_positions
            agent_indices_expanded = agent_indices[:, jnp.newaxis]
            agent_positions_expanded = state.agent_positions[:, jnp.newaxis]

            # Create a matrix of other agent indices for each agent
            other_agent_indices = agent_indices_expanded + jnp.arange(1, num_agents)
            other_agent_indices = other_agent_indices + (other_agent_indices >= agent_indices_expanded)

            # Use other_agent_indices to index into agent_positions
            other_agent_positions = jnp.take(state.agent_positions, other_agent_indices, axis=0)

            obs = vmap_mini_obs(state.agent_positions, agent_indices, other_agent_indices, other_agent_positions)
            pickups = jnp.sum(state.agent_inventories, axis=1) > INTERACT_THRESHOLD
            def agent2show(inventory, freeze):
                agent_to_show = jnp.where(
                    freeze >= 0, inventory, 0
                )
                return agent_to_show
            vmap_agent2show = jax.vmap(agent2show, (0, 0), (0))
            agents2show = vmap_agent2show(state.agent_inventories, state.agent_freezes)
            return {
                "observations": obs,
                "inventory": jnp.array(
                    [
                        state.agent_inventories[:, 0],
                        state.agent_inventories[:, 1],
                        pickups,
                        agents2show[:, 0],
                        agents2show[:, 1],
                    ],
                    dtype=jnp.int8,
                ),
            }

        def _get_reward(state, pair) -> jnp.ndarray:
            inv1 = state.agent_inventories[pair[0]] / state.agent_inventories[pair[0]].sum()
            inv2 = state.agent_inventories[pair[1]] / state.agent_inventories[pair[1]].sum()
            r1 = inv1 @ payoff_matrix[0] @ inv2.T
            r2 = inv1 @ payoff_matrix[1] @ inv2.T
            return jnp.array([r1, r2])


        def _interact(state: State, actions: jnp.array, rng_key: jnp.ndarray) -> Tuple[jnp.ndarray, State]:
            interact_idx = jnp.int8(Items.interact)

            # Remove old interacts from the grid
            grid = jnp.where(
                state.grid == interact_idx, jnp.int8(Items.empty), state.grid
            )
            state = state.replace(grid=grid)

            # Get agent positions and orientations
            agent_positions = state.agent_positions[:, :2]
            agent_orientations = state.agent_positions[:, 2]

            zaps = actions == Actions.interact
            # print(zaps, 'zaps')
            def get_target(agent_pos, grid):
                target = jnp.clip(
                    agent_pos + STEP[agent_pos[2]], 0, GRID_SIZE - 1
                )
                interact = (
                    jnp.isin(grid[target[0], target[1]], AGENT_IDX)
                )
                target_ahead = jnp.clip(
                    agent_pos + 2 * STEP[agent_pos[2]], 0, GRID_SIZE - 1
                )
                interact_ahead = (
                    jnp.isin(grid[target_ahead[0], target_ahead[1]], AGENT_IDX)
                )
                target_right = (
                    agent_pos
                    + STEP[agent_pos[2]]
                    + STEP[(agent_pos[2] + 1) % 4]
                )  
                oob = jnp.logical_or(
                    (target_right > GRID_SIZE - 1).any(),
                    (target_right < 0).any(),
                )
                target_right = jnp.where(oob, target, target_right)

                interact_right = (
                jnp.isin(grid[target_right[0], target_right[1]], AGENT_IDX)
                )
                target_left = (
                    agent_pos
                    + STEP[agent_pos[2]]
                    + STEP[(agent_pos[2] - 1) % 4]
                )
                oob = jnp.logical_or(
                    (target_left > GRID_SIZE - 1).any(),
                    (target_left < 0).any(),
                )
                target_left = jnp.where(oob, target, target_left)
                interact_left = (
                jnp.isin(grid[target_left[0], target_left[1]], AGENT_IDX)
                )
                interact = jnp.logical_or(
                    interact,
                    jnp.logical_or(
                        interact_ahead,
                        jnp.logical_or(interact_right, interact_left),
                    ),
                )
                interact_targets = jnp.array([grid[target[0], target[1]],
                                              grid[target_ahead[0], target_ahead[1]],
                                              grid[target_right[0], target_right[1]],
                                              grid[target_left[0], target_left[1]]])
                return interact, target, target_ahead, target_right, target_left, interact_targets
            
            vmap_get_target = jax.vmap(get_target, (0, None), (0, 0, 0, 0, 0, 0))
            agent_positions, target, target_ahead, target_right, target_left, interact_targets = vmap_get_target(state.agent_positions, state.grid)
            # print(target, target_ahead, target_right, target_left, interact_targets,'target, target_ahead, target_right, target_left, interact_targets')

            def update_grid_interact(i, val):
                grid, target, target_ahead, target_right, target_left, zaps = val
                t_i, t_a_i, t_r_i, t_l_i = target[i], target_ahead[i], target_right[i], target_left[i]
                aux_grid = jnp.copy(grid)
                item = jnp.where(
                    grid[t_i[0], t_i[1]],
                    grid[t_i[0], t_i[1]],
                    interact_idx,
                )
                aux_grid = aux_grid.at[t_i[0], t_i[1]].set(item)

                item = jnp.where(
                    grid[t_a_i[0], t_a_i[1]],
                    grid[t_a_i[0], t_a_i[1]],
                    interact_idx,
                )
                aux_grid = aux_grid.at[
                    t_a_i[0], t_a_i[1]
                ].set(item)

                item = jnp.where(
                    grid[t_r_i[0], t_r_i[1]],
                    grid[t_r_i[0], t_r_i[1]],
                    interact_idx,
                )
                aux_grid = aux_grid.at[
                    t_r_i[0], t_r_i[1]
                ].set(item)

                item = jnp.where(
                    grid[t_l_i[0], t_l_i[1]],
                    grid[t_l_i[0], t_l_i[1]],
                    interact_idx,
                )
                aux_grid = aux_grid.at[t_l_i[0], t_l_i[1]].set(
                    item
                )
                grid = jnp.where(zaps[i], aux_grid, grid)
                return grid, target, target_ahead, target_right, target_left, zaps

            grid, _, _, _, _, _ = jax.lax.fori_loop(0, num_agents, update_grid_interact, (state.grid, target, target_ahead, target_right, target_left, zaps))
            state = state.replace(grid=grid)
            # Create a function to generate agent_pairs without itertools
            def generate_agent_pairs(num_agents):
                pairs = jnp.array([(i, j) for i in range(num_agents) for j in range(i + 1, num_agents)])
                return pairs

            agent_pairs = jnp.array(generate_agent_pairs(num_agents))
            # print(agent_pairs.shape, 'agent pairs')

            # Sort the agent pairs randomly using the provided rng_key
            random_key, subkey = jax.random.split(rng_key)
            shuffled_indices = jax.random.permutation(subkey, jnp.arange(agent_pairs.shape[0]))

            # Create a boolean array to track whether each agent is in an interaction
            agent_interaction_mask = jnp.zeros((num_agents,), dtype=bool)
            pairwise_mask = jnp.zeros((agent_pairs.shape[0],), dtype=bool)

            def process_agent_interactions(i, val):
                agent_interaction_mask, shuffled_indices, interact_targets, agent_inventories, pairwise_mask, zaps = val
                pair_idx = shuffled_indices[i]
                agent0, agent1 = agent_pairs[pair_idx]
                agent1_in_agent0 = jnp.isin(agent1+1, interact_targets[agent0])
                agent0_in_agent1 = jnp.isin(agent0+1, interact_targets[agent1])
                pickup0 = agent_inventories[agent0].sum() > INTERACT_THRESHOLD
                pickup1 = agent_inventories[agent1].sum() > INTERACT_THRESHOLD
                is_interacting = jnp.logical_and(agent1_in_agent0, agent0_in_agent1)
                is_interacting = jnp.logical_and(is_interacting, jnp.logical_and(pickup0, pickup1))
                is_interacting = jnp.logical_and(is_interacting, zaps[agent0])
                is_interacting = jnp.logical_and(is_interacting, zaps[agent1])
                interacting = jnp.logical_and(is_interacting, jnp.logical_not(agent_interaction_mask[agent0] | agent_interaction_mask[agent1]))
                interacting0 = jnp.logical_or(is_interacting, agent_interaction_mask[agent0])
                interacting1 = jnp.logical_or(is_interacting, agent_interaction_mask[agent1])
                agent_interaction_mask = agent_interaction_mask.at[agent0].set(interacting0)
                agent_interaction_mask = agent_interaction_mask.at[agent1].set(interacting1)
                pairwise_mask = pairwise_mask.at[pair_idx].set(interacting)

                return (agent_interaction_mask,  shuffled_indices, interact_targets, agent_inventories, pairwise_mask, zaps)
            
            # for agent0,agent1 in [(0,1),(0,2),(1,2)]:
            #     agent1_in_agent0 = jnp.isin(agent1+1, interact_targets[agent0])
            #     agent0_in_agent1 = jnp.isin(agent0+1, interact_targets[agent1])
            #     pickup0 = state.agent_inventories[agent0].sum() > INTERACT_THRESHOLD
            #     pickup1 = state.agent_inventories[agent1].sum() > INTERACT_THRESHOLD

            #     print('##')
            #     print(agent0, agent1, 'agent0, agent1:')
            #     print(agent1_in_agent0, 'agent1_in_agent0')
            #     print(agent0_in_agent1, 'agent0_in_agent1')
            #     print(pickup0, 'pickup0')
            #     print(pickup1, 'pickup1')
            #     print('##')



            agent_interaction_mask, _, _, _,pairwise_mask, _ = jax.lax.fori_loop(0, num_agents, process_agent_interactions, 
                                                          (agent_interaction_mask,
                                                           shuffled_indices,
                                                           interact_targets,
                                                           state.agent_inventories, pairwise_mask, zaps))


            # def get_interaction_mask(pair, agent_interaction_mask):
            #     i, j = pair[0], pair[1]
            #     return agent_interaction_mask[i] & agent_interaction_mask[j]

            # interaction_masks = jax.vmap(get_interaction_mask, in_axes=(0, None))(agent_pairs, agent_interaction_mask)
            interaction_masks = pairwise_mask
            # print(agent_interaction_mask, 'agent interaction mask')
            # print(interaction_masks, 'interaction masks')
            # print(interaction_masks.shape, 'interaction mask shape')
            # selected_interactions = interaction_masks.all(axis=-1)
            # print(selected_interactions.shape, 'selected_interactions shape')

            def compute_rewards_for_pair(state, pair):
                return _get_reward(state, pair)  # Assume _get_reward is defined elsewhere

            rewards_for_pairs = jax.vmap(compute_rewards_for_pair, in_axes=(None, 0))(state, agent_pairs)

            def get_selected_pairs(interaction_masks,  agent_pairs, rewards_for_pairs):
                selected_pairs = jnp.where(interaction_masks==1, agent_pairs, -1*jnp.ones((2,), dtype=jnp.int8))
                selected_rewards = jnp.where(interaction_masks==1, rewards_for_pairs, 0)
                return  selected_pairs, selected_rewards
            
            
            vmap_get_selected_pairs = jax.vmap(get_selected_pairs, (0,0,0), (0,0))
            selected_pairs, selected_rewards = vmap_get_selected_pairs(interaction_masks, agent_pairs, rewards_for_pairs)
            # print(selected_pairs.shape, 'selected pairs shape')

            rewards = jnp.zeros((num_agents,))
            # print(rewards_for_pairs.shape,'rewards for pairs')
            def assign_reward(i, val):
                rewards, agent_pairs, selected_rewards = val
                agent0, agent1 = agent_pairs[i]
                rewards = rewards.at[agent0].set(selected_rewards[i][0])
                rewards = rewards.at[agent1].set(selected_rewards[i][1])
                return rewards, agent_pairs, selected_rewards
            
            rewards, _, _ = jax.lax.fori_loop(0, selected_rewards.shape[0], assign_reward, (rewards, agent_pairs, selected_rewards))
            # rewards = rewards.at[selected_pairs.ravel()].set(rewards_for_pairs.ravel())
            # print(rewards.shape, 'rewards shape')
            # Update the grid with the selected interact positions
            # def update_grid(i, val):
            #     grid, selected_interact_positions, interact_idx = val
            #     grid = grid.at[selected_interact_positions[i, 0, 0], selected_interact_positions[i, 0, 1]].set(interact_idx)
            #     grid = grid.at[selected_interact_positions[i, 1, 0], selected_interact_positions[i, 1, 1]].set(interact_idx)
            #     return grid, selected_interact_positions, interact_idx
            
            # grid, _, _ = jax.lax.fori_loop(0, selected_interact_positions.shape[0], update_grid, (state.grid, selected_interact_positions, interact_idx))

            # Update state
            state = state.replace(grid=grid)
            def update_interacted_agents(i, val):
                interacted_agents, pairs = val
                # IDFK IF THIS WORKS. DOES IT AUTOMATICALLY TURN IT INTO BOOL???
                interacted_agents = interacted_agents.at[pairs[i, 0], pairs[i, 1]].set(selected_pairs[i,0])
                interacted_agents = interacted_agents.at[pairs[i, 1], pairs[i, 0]].set(selected_pairs[i,0])
                return interacted_agents, pairs

            # Initialize an interacted_agents matrix with False values
            interacted_agents = jnp.zeros((num_agents, num_agents), dtype=bool)

            # Update interacted_agents matrix using lax.scan and selected_pairs
            interacted_agents, _ = jax.lax.fori_loop(0, selected_pairs.shape[0], update_interacted_agents, (interacted_agents, selected_pairs))
            # print(interacted_agents, 'interacted agents')

            return rewards, state, agent_interaction_mask


        def _step(
            key: chex.PRNGKey,
            state: State,
            actions: Tuple[int, int],
        ):

            """Step the environment."""
            actions = jnp.array([actions[i] for i in self.agents])
            original_agent_pos = state.agent_positions
            # print(original_agent_pos, 'original agent pos')
            # freeze check
            # action_0, action_1 = actions
            # action_0 = jnp.where(state.freeze > 0, Actions.stay, action_0)
            # action_1 = jnp.where(state.freeze > 0, Actions.stay, action_1)
            def freeze_check(action, freeze):
                action = jnp.where(freeze > 0, Actions.stay, action)
                return action
            vmap_freeze_check = jax.vmap(freeze_check, (0, 0), 0)
            actions = vmap_freeze_check(jnp.array(actions), state.agent_freezes)
            def turning_and_moving(agent_pos, action):
                # turning red
                new_pos = jnp.int8(
                    (agent_pos + ROTATIONS[action])
                    % jnp.array([GRID_SIZE + 1, GRID_SIZE + 1, 4])
                )

                # get new positions
                # moving red 
                move = action == Actions.forward
                new_pos = jnp.where(
                    move, new_pos + STEP[agent_pos[2]], new_pos
                )
                new_agent_pos = jnp.clip(
                    new_pos,
                    a_min=jnp.array([0, 0, 0], dtype=jnp.int8),
                    a_max=jnp.array(
                        [GRID_SIZE - 1, GRID_SIZE - 1, 3], dtype=jnp.int8
                    ),
                )

                # if you bounced back to ur original space, we change your move to stay (for collision logic)
                move = (new_agent_pos[:2] != agent_pos[:2]).any()
                return new_agent_pos, move

            vmap_turning_and_moving = jax.vmap(turning_and_moving, (0, 0), (0, 0))
            new_agent_positions, moves = vmap_turning_and_moving(state.agent_positions, actions)


            def check_collision(agent1_pos, agent2_pos):
                collision = jnp.all(agent1_pos[:2] == agent2_pos[:2])
                return collision

            #check if there are collisions
            vmap_check_collision = jax.vmap(jax.vmap(check_collision, (None, 0), 0), (0, None), 0)
            collisions = vmap_check_collision(new_agent_positions, new_agent_positions)

            # if collision, priority to whoever didn't move
            # if collision and we are trying to move to a taken spot we stay
            def update_position(new_agent_pos, old_agent_pos, move1, move2, collision):
                new_agent_pos = jnp.where(
                collision
                * move1
                * (1 - move2),  # red moved, blue didn't
                old_agent_pos,
                new_agent_pos,
                )
                return new_agent_pos

            n = len(moves)
            mask = jnp.eye(n, dtype=bool)
            row_selector = ~mask
            move_matrix = jnp.where(row_selector, moves, moves[0])[:, 1:]

            n = collisions.shape[0]
            mask = jnp.eye(n, dtype=bool)
            row_selector = ~mask
            collision_matrix = jnp.where(row_selector, collisions, collisions[0])[:, 1:]
            # print(collision_matrix, 'collision matrix')
            # print(new_agent_positions, 'new agent positions')
            # print(state.agent_positions, 'old agent positions')

            # print('moves', moves)
            # print('move matrix', move_matrix)
            # print('collisions', collisions)
            # print('collision matrix', collision_matrix)
            vmap_update_position = jax.vmap(jax.vmap(update_position, (None, None, None, 0, 0), 0), (0, 0, 0, 0, 0), 0)
            updated_position = vmap_update_position(new_agent_positions, state.agent_positions, moves, move_matrix, collision_matrix)

            def process_positions(positions, old_pos, new_agent_position):
                # Check if all positions along axis 1 are the same.
                same_positions = jnp.all(jnp.all(positions[1:] == positions[:-1], axis=0), axis=-1)

                # Compute the result by selecting either the first position or the predefined value.
                result = jnp.where(same_positions, new_agent_position, old_pos)
                return result
            
            vmap_process_positions = jax.vmap(process_positions, (0, 0, 0), 0)
            updated_position_coll = vmap_process_positions(updated_position, state.agent_positions, new_agent_positions)

            key, subkey = jax.random.split(key)
            choices = jnp.array([0, 1])
            upper_triangle_indices = jnp.triu_indices(n, k=1)
            upper_triangle_values = jax.random.choice(subkey, choices, shape=(n * (n - 1) // 2,))
            lower_triangle_values = (upper_triangle_values + 1 )% 2
            
            upper_triangle_matrix = jnp.zeros((n, n))
            upper_triangle_matrix = upper_triangle_matrix.at[upper_triangle_indices].set(upper_triangle_values)
            lower_triangle_matrix = jnp.zeros((n, n))
            lower_triangle_matrix = lower_triangle_matrix.at[upper_triangle_indices].set(lower_triangle_values) #reusing upper indices and then transposing
            # utm_transposed = upper_triangle_matrix.T + 1 % 2
            
            takes_square_matrix = upper_triangle_matrix + lower_triangle_matrix.T
            # print(takes_square_matrix, 'takes square matrix')
            # print(takes_square_matrix, 'takes square matrix 1')

            n = takes_square_matrix.shape[0]
            mask = jnp.eye(n, dtype=bool)
            row_selector = ~mask
            # whatever = takes_square_matrix[~jnp.eye(takes_square_matrix.shape[0], dtype=bool)].reshape(takes_square_matrix.shape[0], takes_square_matrix.shape[0]-1)
            takes_square_matrix = jnp.where(row_selector, takes_square_matrix, takes_square_matrix[0])[:, 1:]
            # alex1 = takes_square_matrix.reshape(1,takes_square_matrix.shape[0]* takes_square_matrix.shape[0])
            # alex2 = jnp.delete(alex1, jnp.arange(takes_square_matrix.shape[0])*takes_square_matrix.shape[0])
            # alex3 = alex2.reshape(takes_square_matrix.shape[0], takes_square_matrix.shape[0]-1)
            # takes_square_matrix = jnp.where(row_selector, takes_square_matrix, takes_square_matrix[0])
            # upper = takes_square_matrix[jnp.triu_indices(n, k=1)]
            # upper = takes_square_matrix[:, 1:]
            # lower = takes_square_matrix[1:, :takes_square_matrix.shape[0]-1]
            # takes_square_matrix = jnp.concatenate((upper, lower), axis=1)
            # jax.debug.breakpoint()

            # whatever = jnp.delete(takes_square_matrix, jnp.arange(3), axis=0)
            
            # print(takes_square_matrix, 'takes square matrix 2')
            def update_rand_pos(pos_old, pos_new, collision, move1, move2, takes_square):
                new_pos = jnp.where(
                collision
                * move1
                * move2
                * (
                    1 - takes_square
                ),  # if both collide and red doesn't take square
                pos_old,
                pos_new,
                )
                return new_pos
            # print(takes_square_matrix, 'takes square matrix 2')
            vmap_update_rand_pos = jax.vmap(jax.vmap(update_rand_pos, (None, None, 0, None, 0, 0), 0), (0, 0, 0, 0, 0, 0), 0)
            updated_position = vmap_update_rand_pos(state.agent_positions, updated_position_coll, collision_matrix, moves, move_matrix, takes_square_matrix)
            updated_position = vmap_process_positions(updated_position, state.agent_positions, updated_position_coll)
            def update_inventories(new_pos, inventory):
                coop_matches = (
                state.grid[new_pos[0], new_pos[1]] == Items.cooperation_coin
                )
                defect_matches = (
                state.grid[new_pos[0], new_pos[1]] == Items.defection_coin
                )
                inventory = inventory + jnp.array(
                [coop_matches, defect_matches]
                )
                return inventory

            vmap_update_inventory = jax.vmap(update_inventories, (0, 0), 0)

            agent_inventories = vmap_update_inventory(updated_position, state.agent_inventories)

            state = state.replace(agent_inventories=agent_inventories)
            state = state.replace(agent_positions=updated_position)

            key, subkey = jax.random.split(key)
            red_reward, blue_reward = 0, 0
            (
                rewards, 
                state, 
                interacted_agents
            ) = _interact(state, actions, key)
            # print(interacted_agents, 'interacted agents')

            def update_freeze(freeze, interact):
                freeze = jnp.where(
                interact, freeze_penalty, freeze
                )
                return freeze
            vmap_update_freeze = jax.vmap(update_freeze, (0, 0), 0)

            # interact_vector = interacted_agents.any(axis=1)
            state = state.replace(agent_freezes=state.agent_freezes-1)

            agent_freezes = vmap_update_freeze(state.agent_freezes, interacted_agents)
            state = state.replace(agent_freezes=agent_freezes)

            # state_sft_re = _soft_reset_state(key, state)
            key, subkey = jax.random.split(key)

            agent_pos = jax.random.choice(
                subkey, AGENT_SPAWNS, shape=(), replace=False
            )
            player_dir = jax.random.randint(
                subkey, shape=(num_agents,), minval=0, maxval=3, dtype=jnp.int8
            )
            sft_re_player_pos = jnp.array(
                [agent_pos[:num_agents, 0], agent_pos[:num_agents, 1], player_dir]
            ).T

            # print(sft_re_player_pos, 'sft_re_player_pos')
            def update_positions_freezes(new_player_pos, agent_positions, agent_freezes, 
                                        agent_inventories,):
                agent_positions = jnp.where(
                                        agent_freezes==0,
                                        new_player_pos,
                                        agent_positions   
                                    )
                agent_inventories = jnp.where(
                                        agent_freezes==0,
                                        jnp.zeros((2,), dtype=jnp.int8),
                                        agent_inventories)
                agent_freezes = jnp.where(
                                        agent_freezes==0,
                                        -1,
                                        agent_freezes   
                                    )
                return agent_positions, agent_freezes, agent_inventories
            
            vmap_update_positions_freezes = jax.vmap(update_positions_freezes, (0, 0, 0, 0), (0,0,0))
            agent_positions, agent_freezes, agent_inventories = vmap_update_positions_freezes(sft_re_player_pos, 
                                                                state.agent_positions, state.agent_freezes, state.agent_inventories)
            # print(agent_freezes, 'agent freezes')
            # print(agent_inventories, 'agent inventories')
            # print(agent_positions, 'agent positions')

            # print(agent_positions, 'agent positions after freezes')

            def pos_remove(i, val):
                grid, old_pos = val
                grid = grid.at[
                    (old_pos[i, 0], old_pos[i,1])
                ].set(jnp.int8(Items.empty))
                return (grid, old_pos)

            def pos_add(i, val):
                grid, new_pos = val
                grid = grid.at[(new_pos[i,0], new_pos[i,1])].set(
                    jnp.int8(i+1)
                )
                return (grid, new_pos)

            grid, _ = jax.lax.fori_loop(0, num_agents, pos_remove, (state.grid, original_agent_pos))
            grid, _ = jax.lax.fori_loop(0, num_agents, pos_add, (grid, agent_positions))

            state = state.replace(coin_timer_coop=state.coin_timer_coop-1)
            state = state.replace(coin_timer_defect=state.coin_timer_defect-1)

            # define jax foriloop function to update grid cell to coin value when cointimer hits 0 but only if that cell is 0, otherwise take cell value
            def update_grid_coin_coop(i, val):
                grid, coin_timer_coop, coop_coin = val
                cond = jnp.logical_and(coin_timer_coop[i] == 0, grid[coop_coin[i, 0], coop_coin[i, 1]] == 0)
                update_value = jnp.where(cond, Items.cooperation_coin, grid[coop_coin[i, 0], coop_coin[i, 1]])
                updated_grid = grid.at[coop_coin[i, 0], coop_coin[i, 1]].set(update_value)
                return updated_grid, coin_timer_coop, coop_coin

            def update_grid_coin_defect(i, val):
                grid, coin_timer_defect, defect_coin = val
                cond = jnp.logical_and(coin_timer_defect[i] == 0, grid[defect_coin[i, 0], defect_coin[i, 1]] == 0)
                update_value = jnp.where(cond, Items.defection_coin, grid[defect_coin[i, 0], defect_coin[i, 1]])
                updated_grid = grid.at[defect_coin[i, 0], defect_coin[i, 1]].set(update_value)
                return updated_grid, coin_timer_defect, defect_coin

            # now write the foriloop for the above functions
            grid, _, _ = jax.lax.fori_loop(0, NUM_COINS, update_grid_coin_coop, (grid, state.coin_timer_coop, state.coin_coop))
            grid, _, _ = jax.lax.fori_loop(0, NUM_COINS, update_grid_coin_defect, (grid, state.coin_timer_defect, state.coin_defect))


            key, subkey = jax.random.split(key)
            new_coop_coin_timer = jax.random.randint(
                subkey, shape=(NUM_COINS,), minval=3, maxval=10, dtype=jnp.int8
            )
            key, subkey = jax.random.split(key)
            new_defect_coin_timer = jax.random.randint(
                subkey, shape=(NUM_COINS,), minval=3, maxval=10, dtype=jnp.int8
            )

            def update_timers(coop_coin_timer, defect_coin_timer, new_coop_coin_timer, new_defect_coin_timer):
                coop_coin_timer = jnp.where(coop_coin_timer==0, new_coop_coin_timer, coop_coin_timer)
                defect_coin_timer = jnp.where(defect_coin_timer==0, new_defect_coin_timer, defect_coin_timer)
                return coop_coin_timer, defect_coin_timer

            vmap_update_timers = jax.vmap(update_timers, (0, 0, 0, 0), (0, 0))
            coin_timer_coop, coin_timer_defect = vmap_update_timers(state.coin_timer_coop, state.coin_timer_defect, new_coop_coin_timer, new_defect_coin_timer)

            state = state.replace(agent_freezes=agent_freezes)
            state = state.replace(agent_positions=agent_positions)
            state = state.replace(grid=grid)
            state = state.replace(agent_inventories=agent_inventories)
            state = state.replace(coin_timer_coop=coin_timer_coop)
            state = state.replace(coin_timer_defect=coin_timer_defect)

            state_nxt = State(
                agent_positions=state.agent_positions,
                inner_t=state.inner_t + 1,
                outer_t=state.outer_t,
                grid=state.grid,
                agent_inventories=state.agent_inventories,
                coin_coop=state.coin_coop,
                coin_defect=state.coin_defect,
                agent_freezes=state.agent_freezes,
                coin_timer_coop=state.coin_timer_coop,
                coin_timer_defect=state.coin_timer_defect,
            )

            # now calculate if done for inner or outer episode
            inner_t = state_nxt.inner_t
            outer_t = state_nxt.outer_t
            reset_inner = inner_t == num_inner_steps

            # # # if inner episode is done, return start state for next game
            state_re = _reset_state(key)
            state_re = state_re.replace(outer_t=outer_t + 1)
            state = jax.tree.map(
                lambda x, y: jax.lax.select(reset_inner, x, y),
                state_re,
                state_nxt,
            )
            outer_t = state.outer_t
            reset_outer = outer_t == num_outer_steps
            done = {}
            done["__all__"] = reset_outer

            obs = _get_obs(state)
            rewards = jnp.where(reset_inner, 0, rewards)

            return (
                obs,
                state,
                rewards,
                done,
                {"discount": jnp.zeros((), dtype=jnp.int8)},
            )

        def _reset_state(
            key: jnp.ndarray,
        ) -> Tuple[jnp.ndarray, State]:
            key, subkey = jax.random.split(key)

            # coin_pos = jax.random.choice(
            #     subkey, COIN_SPAWNS, shape=(NUM_COIN_TYPES*NUM_COINS,), replace=False
            # )

            agent_pos = jax.random.choice(
                subkey, AGENT_SPAWNS, shape=(), replace=False
            )
            player_dir = jax.random.randint(
                subkey, shape=(num_agents,), minval=0, maxval=3, dtype=jnp.int8
            )
            player_pos = jnp.array(
                [agent_pos[:num_agents, 0], agent_pos[:num_agents, 1], player_dir]
            ).T

            grid = jnp.zeros((GRID_SIZE, GRID_SIZE), jnp.int8)

            def pos_fun(i, val):
                grid, player_pos = val
                grid = grid.at[player_pos[i, 0], player_pos[i, 1]].set(
                jnp.int8(i+1))
                return (grid, player_pos)

            grid, _ = jax.lax.fori_loop(0, num_agents, pos_fun, (grid, player_pos))

            if fixed_coin_location:
                rand_idx = jax.random.randint(
                    subkey, shape=(), minval=0, maxval=1
                )
                red_coins = jnp.where(rand_idx, RED_SPAWN, BLUE_SPAWN)
                blue_coins = jnp.where(rand_idx, BLUE_SPAWN, RED_SPAWN)
            else:
                coin_spawn = jax.random.permutation(
                    subkey, COIN_SPAWNS, axis=0
                )
                red_coins = coin_spawn[:NUM_COINS, :]
                blue_coins = coin_spawn[NUM_COINS:, :]

            for i in range(NUM_COINS):
                grid = grid.at[red_coins[i, 0], red_coins[i, 1]].set(
                    jnp.int8(Items.defection_coin)
                )

            for i in range(NUM_COINS):
                grid = grid.at[blue_coins[i, 0], blue_coins[i, 1]].set(
                    jnp.int8(Items.cooperation_coin)
                )
            key, subkey = jax.random.split(key)
            #initialize coin timer to random numbers between 1 and 10
            coin_timer_coop = jax.random.randint(
                subkey, shape=(NUM_COINS,), minval=3, maxval=10
            )
            key, subkey = jax.random.split(key)
            coin_timer_defect = jax.random.randint(
                subkey, shape=(NUM_COINS,), minval=3, maxval=10
            )

            return State(
                agent_positions=player_pos,
                inner_t=0,
                outer_t=0,
                grid=grid,
                agent_inventories=jnp.zeros((num_agents, 2)),
                coin_defect=red_coins,
                coin_coop=blue_coins,
                agent_freezes=jnp.int16(-1*jnp.ones((num_agents,))),
                coin_timer_coop=coin_timer_coop,
                coin_timer_defect=coin_timer_defect,
            )

        def reset(
            key: jnp.ndarray,
        ) -> Tuple[jnp.ndarray, State]:
            state = _reset_state(key)
            obs = _get_obs(state)
            return obs, state

        # overwrite Gymnax as it makes single-agent assumptions
        # self.step = jax.jit(_step)
        self.step_env = _step
        self.reset = jax.jit(reset)
        self.get_obs_point = _get_obs_point
        self.get_reward = _get_reward

        # for debugging
        self.get_obs = _get_obs
        self.cnn = True

        self.num_inner_steps = num_inner_steps
        self.num_outer_steps = num_outer_steps
        self.num_agents = num_agents
        _shape = (
            (OBS_SIZE, OBS_SIZE, len(Items) - 1 + 4)
            if self.cnn
            else (OBS_SIZE**2 * (len(Items) - 1 + 4),)
        )
        self.observation_spaces = {
            i: {"observation": spaces.Box(
                low=0, high=1, shape=_shape, dtype=jnp.uint8),
            "inventory": spaces.Box(
                low=0,high=NUM_COINS,shape=NUM_COIN_TYPES + 4,dtype=jnp.uint8,),
        } for i in range(self.num_agents)}
        self.action_spaces = {
            i: spaces.Discrete(len(Actions)) for i in range(self.num_agents)
        }

    @property
    def name(self) -> str:
        """Environment name."""
        return "MGinTheGrid"

    @property
    def num_actions(self) -> int:
        """Number of actions possible in environment."""
        return len(Actions)


    def state_space(self) -> spaces.Dict:
        """State space of the environment."""
        _shape = (
            (GRID_SIZE, GRID_SIZE, NUM_TYPES + 4)
            if self.cnn
            else (GRID_SIZE**2 * (NUM_TYPES + 4),)
        )
        return spaces.Box(low=0, high=1, shape=_shape, dtype=jnp.uint8)

    @classmethod
    def render_tile(
        cls,
        obj: int,
        agent_dir: Union[int, None] = None,
        agent_hat: bool = False,
        highlight: bool = False,
        tile_size: int = 32,
        subdivs: int = 3,
    ) -> onp.ndarray:
        """
        Render a tile and cache the result
        """

        # Hash map lookup key for the cache
        key: tuple[Any, ...] = (agent_dir, agent_hat, highlight, tile_size)
        if obj:
            key = (obj, 0, 0, 0) + key if obj else key

        if key in cls.tile_cache:
            return cls.tile_cache[key]

        img = onp.zeros(
            shape=(tile_size * subdivs, tile_size * subdivs, 3),
            dtype=onp.uint8,
        )

        # Draw the grid lines (top and left edges)
        fill_coords(img, point_in_rect(0, 0.031, 0, 1), (100, 100, 100))
        fill_coords(img, point_in_rect(0, 1, 0, 0.031), (100, 100, 100))
    # class Items(IntEnum):

        if obj == Items.agent1:
            # Draw the agent 1
            agent_color = PLAYER1_COLOUR
        elif obj == Items.agent2:
            # Draw agent 2
            agent_color = PLAYER2_COLOUR
        elif obj == Items.agent3:
            # Draw agent 3
            agent_color = PLAYER3_COLOUR
        elif obj == Items.agent4:
            # Draw agent 4
            agent_color = PLAYER4_COLOUR
        elif obj == Items.agent5:
            # Draw agent 5
            agent_color = PLAYER5_COLOUR
        elif obj == Items.agent6:
            # Draw agent 6
            agent_color = PLAYER6_COLOUR
        elif obj == Items.agent7:
            # Draw agent 7
            agent_color = PLAYER7_COLOUR
        elif obj == Items.agent8:
            # Draw agent 8
            agent_color = PLAYER8_COLOUR
        elif obj == Items.defection_coin:
            # Draw the red coin as GREEN COOPERATE
            fill_coords(
                img, point_in_circle(0.5, 0.5, 0.31), (44.0, 160.0, 44.0)
            )
        elif obj == Items.cooperation_coin:
            # Draw the blue coin as DEFECT/ RED COIN
            fill_coords(
                img, point_in_circle(0.5, 0.5, 0.31), (214.0, 39.0, 40.0)
            )
        elif obj == Items.wall:
            fill_coords(img, point_in_rect(0, 1, 0, 1), (127.0, 127.0, 127.0))

        elif obj == Items.interact:
            fill_coords(img, point_in_rect(0, 1, 0, 1), (188.0, 189.0, 34.0))

        elif obj == 99:
            fill_coords(img, point_in_rect(0, 1, 0, 1), (44.0, 160.0, 44.0))

        elif obj == 100:
            fill_coords(img, point_in_rect(0, 1, 0, 1), (214.0, 39.0, 40.0))

        elif obj == 101:
            # white square
            fill_coords(img, point_in_rect(0, 1, 0, 1), (255.0, 255.0, 255.0))

        # Overlay the agent on top
        if agent_dir is not None:
            if agent_hat:
                tri_fn = point_in_triangle(
                    (0.12, 0.19),
                    (0.87, 0.50),
                    (0.12, 0.81),
                    0.3,
                )

                # Rotate the agent based on its direction
                tri_fn = rotate_fn(
                    tri_fn,
                    cx=0.5,
                    cy=0.5,
                    theta=0.5 * math.pi * (1 - agent_dir),
                )
                fill_coords(img, tri_fn, (255.0, 255.0, 255.0))

            tri_fn = point_in_triangle(
                (0.12, 0.19),
                (0.87, 0.50),
                (0.12, 0.81),
                0.0,
            )

            # Rotate the agent based on its direction
            tri_fn = rotate_fn(
                tri_fn, cx=0.5, cy=0.5, theta=0.5 * math.pi * (1 - agent_dir)
            )
            fill_coords(img, tri_fn, agent_color)

        # Highlight the cell if needed
        if highlight:
            highlight_img(img)

        # Downsample the image to perform supersampling/anti-aliasing
        img = downsample(img, subdivs)

        # Cache the rendered tile
        cls.tile_cache[key] = img
        return img

    def render(
        self,
        state: State,
    ) -> onp.ndarray:
        """
        Render this grid at a given scale
        :param r: target renderer object
        :param tile_size: tile size in pixels
        """
        tile_size = 32
        highlight_mask = onp.zeros_like(onp.array(GRID))

        # Compute the total grid size
        width_px = GRID.shape[0] * tile_size
        height_px = GRID.shape[0] * tile_size

        img = onp.zeros(shape=(height_px, width_px, 3), dtype=onp.uint8)
        grid = onp.array(state.grid)
        grid = onp.pad(
            grid, ((PADDING, PADDING), (PADDING, PADDING)), constant_values=Items.wall
        )
        for a in range(self.num_agents):
            startx, starty = self.get_obs_point(
                state.agent_positions[a, 0], 
                state.agent_positions[a, 1], 
                state.agent_positions[a, 2]
            )
            highlight_mask[
                startx : startx + OBS_SIZE, starty : starty + OBS_SIZE
            ] = True

        # Render the grid
        for j in range(0, grid.shape[1]):
            for i in range(0, grid.shape[0]):
                cell = grid[i, j]
                if cell == 0:
                    cell = None
                agent_here = []
                for a in range(1, self.num_agents+1):
                    agent_here.append(cell == a)
                # if cell in [1,2]:
                #     print(f'coordinates: {i},{j}')
                #     print(cell)


                agent_dir = None
                for a in range(self.num_agents):
                    agent_dir = (
                        state.agent_positions[a,2].item()
                        if agent_here[a]
                        else agent_dir
                    )
                
                agent_hat = False
                for a in range(self.num_agents):
                    agent_hat = (
                        bool(state.agent_inventories[a].sum() > INTERACT_THRESHOLD)
                        if agent_here[a]
                        else agent_hat
                    )

                tile_img = InTheGrid.render_tile(
                    cell,
                    agent_dir=agent_dir,
                    agent_hat=agent_hat,
                    highlight=highlight_mask[i, j],
                    tile_size=tile_size,
                )

                ymin = j * tile_size
                ymax = (j + 1) * tile_size
                xmin = i * tile_size
                xmax = (i + 1) * tile_size
                img[ymin:ymax, xmin:xmax, :] = tile_img

        img = onp.rot90(
            img[
                (PADDING - 1) * tile_size : -(PADDING - 1) * tile_size,
                (PADDING - 1) * tile_size : -(PADDING - 1) * tile_size,
                :,
            ],
            2,
        )
        # Render the inventory
        # agent_inv = []
        # for a in range(self.num_agents):
        #     agent_inv.append(self.render_inventory(state.agent_inventories[a], img.shape[1]))


        time = self.render_time(state, img.shape[1])
        # img = onp.concatenate((img, *agent_inv, time), axis=0)
        img = onp.concatenate((img, time), axis=0)
        return img

    def render_inventory(self, inventory, width_px) -> onp.array:
        tile_height = 32
        height_px = NUM_COIN_TYPES * tile_height
        img = onp.zeros(shape=(height_px, width_px, 3), dtype=onp.uint8)
        tile_width = width_px // NUM_COINS
        for j in range(0, NUM_COIN_TYPES):
            num_coins = inventory[j]
            for i in range(int(num_coins)):
                cell = None
                if j == 0:
                    cell = 99
                elif j == 1:
                    cell = 100
                tile_img = InTheGrid.render_tile(cell, tile_size=tile_height)
                ymin = j * tile_height
                ymax = (j + 1) * tile_height
                xmin = i * tile_width
                xmax = (i + 1) * tile_width
                img[ymin:ymax, xmin:xmax, :] = onp.resize(
                    tile_img, (tile_height, tile_width, 3)
                )
        return img

    def render_time(self, state, width_px) -> onp.array:
        inner_t = state.inner_t
        outer_t = state.outer_t
        tile_height = 32
        img = onp.zeros(shape=(2 * tile_height, width_px, 3), dtype=onp.uint8)
        tile_width = width_px // (self.num_inner_steps)
        j = 0
        for i in range(0, inner_t):
            ymin = j * tile_height
            ymax = (j + 1) * tile_height
            xmin = i * tile_width
            xmax = (i + 1) * tile_width
            img[ymin:ymax, xmin:xmax, :] = onp.int8(255)
        tile_width = width_px // (self.num_outer_steps)
        j = 1
        for i in range(0, outer_t):
            ymin = j * tile_height
            ymax = (j + 1) * tile_height
            xmin = i * tile_width
            xmax = (i + 1) * tile_width
            img[ymin:ymax, xmin:xmax, :] = onp.int8(255)
        return img





"""
Introduction running and visualising the MPE environments using random actions.
"""

import jax 
from jaxmarl import make
from jaxmarl.environments.mpe import MPEVisualizer

# Parameters + random keys
max_steps = 25
key = jax.random.PRNGKey(0)
key, key_r, key_a = jax.random.split(key, 3)

# Instantiate environment
env = make('MPE_simple_reference_v3')
obs, state = env.reset(key_r)
print('list of agents in environment', env.agents)

# Sample random actions
key_a = jax.random.split(key_a, env.num_agents)
actions = {agent: env.action_space(agent).sample(key_a[i]) for i, agent in enumerate(env.agents)}
print('example action dict', actions)


state_seq = []
for _ in range(max_steps):
    state_seq.append(state)
    # Iterate random keys and sample actions
    key, key_s, key_a = jax.random.split(key, 3)
    key_a = jax.random.split(key_a, env.num_agents)
    actions = {agent: env.action_space(agent).sample(key_a[i]) for i, agent in enumerate(env.agents)}

    # Step environment
    obs, state, rewards, dones, infos = env.step(key_s, state, actions)


viz = MPEVisualizer(env, state_seq)
viz.animate(view=True)


"""
In this file is defined a simple OBL agent that aligns with the original torch version and
can be used with the (flax-translated) OBL original weights.

This implementation had the goal to perform inference with pretrained params, for this reason
is kept as minimal as possible.
"""

import jax
from jax import numpy as jnp
import flax.linen as nn
from functools import partial
from typing import Tuple
from chex import Array, PRNGKey
from flax.linen.module import compact, nowrap

class MultiLayerLSTM(nn.RNNCellBase):

    num_layers: int
    features: int

    @compact
    def __call__(self, carry, inputs):

        new_hs = []
        new_cs = []
        for l in range(self.num_layers):
            new_carry, y = nn.LSTMCell(self.features, name=f"l{l}")(
                jax.tree.map(lambda x: x[l], carry), inputs
            )
            new_cs.append(new_carry[0])
            new_hs.append(new_carry[1])
            inputs = y

        new_final_carry = (jnp.stack(new_cs), jnp.stack(new_hs))
        return new_final_carry, y

    @nowrap
    def initialize_carry(
        self, rng: PRNGKey, batch_dims: Tuple[int, ...]
    ) -> Tuple[Array, Array]:
        mem_shape = (self.num_layers,) + batch_dims + (self.features,)
        c = jnp.zeros(mem_shape)
        h = jnp.zeros(mem_shape)
        return (c, h)

    @property
    def num_feature_axes(self) -> int:
        return 1


class OBLAgentR2D2(nn.Module):

    hid_dim: int = 512
    out_dim: int = 21
    num_lstm_layer: int = 2
    num_ff_layer: int = 1

    @compact
    def __call__(self, carry, inputs):

        priv_s, publ_s = inputs

        # private net
        priv_o = nn.Sequential(
            [
                nn.Dense(self.hid_dim, name="priv_net_dense_0"),
                nn.relu,
                nn.Dense(self.hid_dim, name="priv_net_dense_1"),
                nn.relu,
                nn.Dense(self.hid_dim, name="priv_net_dense_2"),
                nn.relu,
            ]
        )(priv_s)

        # public net (MLP+lstm)
        x = nn.Sequential([nn.Dense(self.hid_dim, name="publ_net_dense_0"), nn.relu])(
            publ_s
        )
        carry, publ_o = MultiLayerLSTM(
            num_layers=self.num_lstm_layer, features=self.hid_dim, name="lstm"
        )(carry, x)

        o = priv_o * publ_o

        a = nn.Dense(self.out_dim, name="fc_a")(o)

        return carry, a

    @partial(jax.jit, static_argnums=[0])
    def greedy_act(self, params, carry, inputs):

        obs, legal_move = inputs
        priv_s = obs
        publ_s = obs[..., 125:]

        carry, adv = self.apply(params, carry, (priv_s, publ_s))

        legal_adv = (1 + adv - adv.min()) * legal_move
        greedy_action = jnp.argmax(legal_adv, axis=-1)

        return carry, greedy_action

    @nowrap
    def initialize_carry(
        self, rng: PRNGKey, batch_dims: Tuple[int, ...]
    ) -> Tuple[Array, Array]:
        return MultiLayerLSTM(
            num_layers=self.num_lstm_layer, features=self.hid_dim
        ).initialize_carry(rng, batch_dims)



def example():

    from jaxmarl import make
    from jaxmarl.wrappers.baselines import load_params

    weight_file = "./obl-r2d2-flax/icml_OBL1/OFF_BELIEF1_SHUFFLE_COLOR0_BZA0_BELIEF_a.safetensors"
    params = load_params(weight_file)

    agent = OBLAgentR2D2()
    agent_carry = agent.initialize_carry(jax.random.PRNGKey(0), batch_dims=(2,))
    
    rng = jax.random.PRNGKey(0)
    env = make('hanabi')
    obs, env_state = env.reset(rng)
    env.render(env_state)

    batchify = lambda x: jnp.stack([x[agent] for agent in env.agents])
    unbatchify = lambda x: {agent:x[i] for i, agent in enumerate(env.agents)}

    agent_input = (
        batchify(obs),
        batchify(env.get_legal_moves(env_state))
    )
    agent_carry, actions = agent.greedy_act(params, agent_carry, agent_input)
    actions = unbatchify(actions)

    obs, env_state, rewards, done, info = env.step(rng, env_state, actions)

    print('actions:', {agent:env.action_encoding[int(a)] for agent, a in actions.items()})
    env.render(env_state)

if __name__ == "__main__":
    example()





from .grid_map import GridMapCircleAgents, GridMapPolygonAgents, GridMapBarn, GridMapPolygonAgentsSingleMap, GridMapFromBuffer
from .map import Map

def make_map(map_id: str, num_agents: int, rad: float, **map_kwargs) -> GridMapCircleAgents:  # note this type hint technically should be Map

    if map_id not in registered_maps:
        raise ValueError(f"Map: {map_id} not registered!")
    if map_id == "Grid-Rand":
        return GridMapCircleAgents(num_agents=num_agents, rad=rad, **map_kwargs)
    if map_id == "Grid-Rand-Poly":
        return GridMapPolygonAgents(num_agents=num_agents, rad=rad, **map_kwargs)
    if map_id == "Grid-Rand-Barn":
        return GridMapBarn(num_agents=num_agents, rad=rad, **map_kwargs)
    if map_id == "Grid-Rand-Poly-Single":
        return GridMapPolygonAgentsSingleMap(num_agents=num_agents, rad=rad, **map_kwargs)
    if map_id == "Grid-Buffer":
        return GridMapFromBuffer(num_agents=num_agents, rad=rad, **map_kwargs)

registered_maps = [
    "Grid-Rand",
    "Grid-Rand-Poly",
    "Grid-Rand-Barn",
    "Grid-Rand-Poly-Single",
    "Grid-Buffer",
]


import jax
from jax import numpy as jnp
import numpy as np
from functools import partial
import matplotlib.axes._axes as axes
import chex 

class Map(object):
    """ Base class for a map """
    
    def __init__(
        self,
        num_agents,
        rad,
        map_size,
        start_pad=1.5,
        valid_path_check=False,
    ):
        assert start_pad>=1.0, 'start_pad must be greater than or equal to 1.0'
        
        self.num_agents = num_agents
        self.rad = rad
        self.map_size = map_size
        self.start_pad=start_pad
        self.valid_path_check = valid_path_check
        
        # Test case sampling TODO fix this to be not hard coded
        self.dist_to_goal = 50 # 5
        self.rrt_samples = 1000
        self.rrt_step_size = 0.25
        self.goal_radius = 0.3
        
    @partial(jax.jit, static_argnums=[0])
    def sample_scenario(self, key):
        """ Sample map grid and agent start/goal positions """
        
        key_map, key_case = jax.random.split(key)
        map_data = self.sample_map(key_map)
        test_case = self.sample_test_case(key_case, map_data)
        return map_data, test_case
    
    @partial(jax.vmap, in_axes=[None, 0])
    def sample_scenarios(self, key):
        return self.sample_scenario(key)
        
    @partial(jax.jit, static_argnums=[0])
    def sample_map(self, key):
        raise NotImplementedError
    
    @partial(jax.jit, static_argnums=[0])
    def sample_test_case(self, key):
        """ Sample a test case for a given map """
        key, _key = jax.random.split(key)
        map_data = self.sample_map(_key)
        radii = jnp.array([self.rad*self.start_pad, self.goal_radius])

        def _sample_pair(key: chex.PRNGKey):
            """ Sample a start and goal pose for an agent """
            key_s, key_g, key_t = jax.random.split(key, 3)
            low_lim = 1 + self.rad 
            high_lim = self.map_size[1] - 1 - self.rad
            start = jax.random.uniform(key_s, (1, 2), minval=low_lim, maxval=high_lim)
            g_low_lim = jnp.clip(start - self.dist_to_goal, low_lim, high_lim)
            g_high_lim = jnp.clip(start + self.dist_to_goal, low_lim, high_lim)
            goal = jax.random.uniform(key_g, (1, 2), minval=g_low_lim, maxval=g_high_lim)
            theta = jax.random.uniform(key_t, (2, 1), minval=-jnp.pi, maxval=jnp.pi)
            positions = jnp.concatenate([start, goal], axis=0)
            poses = jnp.concatenate([positions, theta], axis=1)
            return poses
        
        def _agent_collision(pos, test_case, rad): 
            dists = jnp.linalg.norm(test_case-pos, axis=1) <= rad*2
            return jnp.any(dists)
        
        def _cond_idx(val):
            """ true while i is less than the number of agents """
            key, i, case = val
            return i < case.shape[0]
        
        def _body_idx(val, key):
            """ samples a start and goal pair for an agent, taking into account pairs sampled previously """
            i, case = val
            def _cond_pos(val):
                """ Check if the sampled pair is valid. Checks
                1. check start and goal do not collide with the map
                3. check the start is not within a radius with other agents' starts
                4. check the start is not within a radius with other agents' starts 
                5. check if start and goal are too close
                """
                
                key, pair, case = val
                temp_case = case.at[i].set(pair+self.rad*3)  # ensure no conflict 
                
                map_collisions = jnp.any(jax.vmap(self.check_circle_map_collision, in_axes=[0, None, 0])(pair[:, :2], map_data, radii))
                agent_collisions = jnp.any(jax.vmap(_agent_collision, in_axes=[0, 1, None])(pair[:, :2], temp_case[:, :, :2], self.rad*self.start_pad))
                
                too_close = (jnp.linalg.norm(pair[0, :2] - pair[1, :2]) <= 2*self.rad).astype(jnp.bool_)
                
                check = map_collisions | agent_collisions | too_close
                
                if self.valid_path_check:
                    valid_path = self.passable_check(pair[0, :2], pair[1, :2], map_data)  # WARNING can make code too slow
                    check = check | ~valid_path
                
                return check
                                
                # return map_collisions | agent_collisions | too_close | ~valid_path
                
                #print('p map', jnp.any(pmap_collision(pair, map_grid, rad)))
                #jax.debug.print('p {pair} map {p}, s ag {s}, g ag {g}, dist f{d} dist {c}', pair=pair, p=jnp.any(pmap_collision(pair, map_grid, rad)), s=agent_collision(pair[0], temp_case[:, 0, :], rad), g=agent_collision(pair[1], temp_case[:, 1, :], rad), d=jnp.linalg.norm(pair[0] - pair[1]) >= dist_to_goal, c=(jnp.linalg.norm(pair[0] - pair[1]) <= 2*rad).astype(jnp.bool_))
                """ true while pos is not valid """
                # 1. check if start collides with map
                # 2. check if goal collides with map
                # 3. check if start collides with other agents
                # 4. check if goal collides with other agents's goals 
                # 5. check if start and goal are too close
                
                return self.check_agent_map_collision(pair[0], map_data, self.rad*self.start_pad) \
                    | self.check_agent_map_collision(pair[1], map_data) \
                    | _agent_collision(pair[0], temp_case[:, 0, :], self.rad) \
                    | _agent_collision(pair[1], temp_case[:, 1, :], self.rad) \
                    | (jnp.linalg.norm(pair[0] - pair[1]) >= self.dist_to_goal).astype(jnp.bool_) \
                    | (jnp.linalg.norm(pair[0] - pair[1]) <= 2*self.rad).astype(jnp.bool_) #\ | ~(self.rrt(key_rrt, map_data, pair[0], pair[1]))
        
            def _body_pos(val):
                """ Sample a start and goal pair """
                key, pair, case = val
                key, key_point = jax.random.split(key)
                pair = _sample_pair(key_point)
                case = case.at[i].set(pair)
                #jax.debug.print('case {c}', c=case)
                return key, pair, case
        
            key, key_point = jax.random.split(key)
            pair = _sample_pair(key_point)
            case = case.at[i].set(pair)
                        
            key, pair, case = jax.lax.while_loop(
                _cond_pos,
                _body_pos,
                (key, pair, case),
            )
            
            i += 1
            return (i, case), None

        fill_max = jnp.max(jnp.array(self.map_size)) + self.rad*2
        case = jnp.full((self.num_agents, 2, 3), fill_max)  # [num_agents, [start_pose, goal_pose]]

        i = 0
        
        key_scan = jax.random.split(key, self.num_agents)
        (_, case), _ = jax.lax.scan(_body_idx, (i, case), key_scan)
        
        # Add intial orientation
        #theta = jax.random.uniform(key, (self.num_agents, 2, 1), minval=-jnp.pi, maxval=jnp.pi)
        #case = jnp.concatenate([case, theta], axis=-1)        
        return map_data, case
    
    @partial(jax.jit, static_argnums=[0])
    def check_circle_map_collision(self, pos, map_data, rad=None):
        """ Check collision between a circle at position `pos` of radius `rad` and the map.
        If rad is None, use the class rad. """
        raise NotImplementedError
    
    @partial(jax.jit, static_argnums=[0])
    def check_agent_map_collision(self, pos, theta, map_data, **agent_kwargs):
        """ Check collision between an agent at position `pos` and the map"""
        # NOTE should we switch these functions to use pose, i.e. [pos_x, pos_y, theta]?
        raise NotImplementedError
    
    def check_all_agent_agent_collisions(self, pos, theta):
        """ Check collision between all agents """
        raise NotImplementedError
        
    def check_agent_beam_intersect(self, beam, pos, theta, range_resolution, **agent_kwargs):
        """ Check for intersection between a lidar beam and an agent. """
        raise NotImplementedError
    
    @partial(jax.jit, static_argnums=[0])
    def check_point_map_collision(self, pos, map_data):
        """ Check collision between `pos` and the map"""
        raise NotImplementedError
    
    @partial(jax.jit, static_argnums=[0])
    def check_line_collision(self, pos1, pos2, map_data):
        """ Check collision between line (pos1) -- (pos2) the map"""
        raise NotImplementedError
    
    def passable_check(self, pos1, pos2, map_data):
        """ Check whether a path exists between pos1 and pos2.
        Note, this does not return the path, only whether a path exists. """
        raise NotImplementedError
    
    @partial(jax.jit, static_argnums=[0])
    def check_agent_translation(self, start, end, map_data):
        """ True for valid translation, False for invalid translation """
        
        l_slope = (end[1] - start[1]) / (end[0] - start[0])
        perpendicular_slope = -1 / l_slope
        delta_x = self.rad / (1 + perpendicular_slope**2)**0.5
        delta_y = perpendicular_slope * delta_x
        #jax.debug.print('per slope {p}', p=perpendicular_slope)
        # Calculate the coordinates of the two points
        delta = jnp.array([delta_x, delta_y])
        start_lower = start + delta
        start_upper = start - delta    
        end_lower = end + delta
        end_upper = end - delta
    
        lower = self.check_line_collision(start_lower, end_lower, map_data)
        upper = self.check_line_collision(start_upper, end_upper, map_data)
        #jax.debug.print('lower {l}, upper {u}', l=lower, u=upper)
        return ~lower & ~upper
    

    @partial(jax.jit, static_argnums=[0])
    def rrt(self, key, map_data, start, goal):
        """ Run RRT algorithm to find a path between start and goal """

        INF = 10000

        print('key')
        low_lim = 1 + self.rad 
        high_lim = self.map_size[1] - 1 - self.rad
        goal_square = jnp.floor(goal).astype(jnp.int32).squeeze()
        #print('goal square', goal_square)
        
        tree = jnp.empty((self.rrt_samples, 3))  # [sample, [x, y, parent]]
        tree = tree.at[0].set(jnp.append(start, -1))
        rrt_idx = 1
        gr = False
            
        def _cond_fun(val):
            i, gr = val[0], val[1]
            return (i < self.rrt_samples) & ~gr
        
        def _body_fun(val):
            i, gr, rrt_idx, tree, key = val
            key, key_s = jax.random.split(key)
            # Sample position, find closest idx and increment towards sampled pos
            sampled_pos = jax.random.uniform(key_s, (2,), minval=low_lim, maxval=high_lim)
            #closest_idx = jnp.argmin(jnp.linalg.norm(tree[:, :2] - sampled_pos, axis=1))
            distances = jnp.linalg.norm(tree[:, :2] - sampled_pos, axis=1)
            distance_mask = jnp.where(jnp.arange(self.rrt_samples) < rrt_idx, 0, 1) * INF
            closest_idx = jnp.argmin(distances + distance_mask)

            tree_pos = tree[closest_idx, :2]
            step_size = jnp.minimum(self.rrt_step_size, jnp.linalg.norm(sampled_pos - tree_pos))
            
            angle = jnp.arctan2(sampled_pos[1] - tree_pos[1], sampled_pos[0] - tree_pos[0])
            test_pos = tree_pos + jnp.array([jnp.cos(angle), jnp.sin(angle)]) * step_size
            
            # Check free space, line collision
            free_space = ~self.check_agent_map_collision(test_pos, map_data)
            line_collision = self.check_agent_translation(tree_pos, test_pos, map_data)
            valid = free_space & line_collision
            
            #goal_square_reached = jnp.array_equal(jnp.floor(test_pos), goal_square)
            goal_reached = jnp.linalg.norm(test_pos - goal) < self.goal_radius
            new_node = jax.lax.select(valid, jnp.concatenate([test_pos, jnp.array([closest_idx])]), jnp.zeros((3,)))
            tree = tree.at[rrt_idx].set(new_node)
            
            rrt_idx += 1*valid
            gr = gr | (goal_reached & valid)
            return (i+1, gr, rrt_idx, tree, key)
            
        val = jax.lax.while_loop(_cond_fun, _body_fun, (0, gr, rrt_idx, tree, key))
        gr, tree = val[1], val[3]    
        
        return tree, gr
    
    @partial(jax.jit, static_argnums=[0])
    def rrt_star(self, key, map_data, start, goal):
        """ Run RRT* algorithm to find an optimal path between start and goal """

        INF = 10000

        low_lim = 1 + self.rad 
        high_lim = self.map_size[1] - 1 - self.rad
        goal_square = jnp.floor(goal).astype(jnp.int32).squeeze()
        
        check_point_connections = jax.vmap(self.check_agent_translation, in_axes=(None, 0, None))
        
        tree = jnp.empty((self.rrt_samples, 5))  # [sample, [x, y, parent, cost, goal]]
        #print('start',start)
        tree = tree.at[0].set(jnp.append(start, jnp.array([-1, 0.0, 0.0])))
        rrt_idx = 1
        goal_reached = False
        goal_idx = jnp.full((30,), -1, dtype=jnp.int32)
            
        rrt_star_neighbours = 20
            
        def _cond_fun(val):
            i, goal_reached = val[0], val[1]
            return (i < self.rrt_samples) # & ~goal_reached
        
        def _body_fun(val):
            #jax.debug.print('body start')
            i, goal_reached, rrt_idx, tree, key = val
            key, key_s = jax.random.split(key)
            #jax.debug.print('rrt idx {r}', r=rrt_idx)
            # Sample position, find closest idx and increment towards sampled pos
            sampled_pos = jax.random.uniform(key_s, (2,), minval=low_lim, maxval=high_lim)
            distances = jnp.linalg.norm(tree[:, :2] - sampled_pos, axis=1)
            distance_mask = jnp.where(jnp.arange(self.rrt_samples) < rrt_idx, 0, 1) * INF

            closest_idx = jnp.argmin(distances + distance_mask)
            #jax.debug.print('closest idx {c}, d {d}', c=closest_idx, d=distances + distance_mask)
            tree_pos = tree[closest_idx, :2]
            step_size = jnp.minimum(self.rrt_step_size, jnp.linalg.norm(sampled_pos - tree_pos))
            
            angle = jnp.arctan2(sampled_pos[1] - tree_pos[1], sampled_pos[0] - tree_pos[0])
            test_pos = tree_pos + jnp.array([jnp.cos(angle), jnp.sin(angle)]) * step_size
            
            # Check free space, line collision
            free_space = ~self.check_agent_map_collision(test_pos, map_data)
            circle_trans = self.check_agent_translation(test_pos, tree_pos, map_data) # NOTE do we need this for our valid check?
            valid = free_space & circle_trans
            #goal_reached = jnp.array_equal(jnp.floor(test_pos), goal_square)
            goal_just_reached = jnp.linalg.norm(test_pos - goal) < self.goal_radius
            #jax.debug.print('valid {v}, free space {f}, circle trans {c}', v=valid, f=free_space, c=circle_trans)

            # Find parent
            tree_dist = jnp.linalg.norm(tree[:, :2] - test_pos, axis=1)  # todo correct for zeros
            distance_mask = jnp.where(jnp.arange(self.rrt_samples) < rrt_idx, 0, 1) * INF
            #jax.debug.print('tree dist {t}', t=tree_dist.shape)
            parent_poss = jnp.argsort(tree_dist+distance_mask)[:rrt_star_neighbours]

            #jax.debug.print('parent poss {p}', p=parent_poss)
            in_range = jnp.where(parent_poss < rrt_idx, True, False)
            invalid_parent = ~check_point_connections(test_pos, tree[parent_poss, :2], map_data) | ~in_range
            #print('invalid parent', invalid_parent)
            #jax.debug.print('invalid parent {i}, not in range {r}', i=invalid_parent, r=~in_range)
            invalid_parent = invalid_parent | ~in_range
            parent_cost = tree[parent_poss, 3] + tree_dist[parent_poss] + INF*invalid_parent
            parent_rel_idx = jnp.argmin(parent_cost)
            parent_idx = parent_poss[parent_rel_idx]
            cost = jnp.min(parent_cost)
            valid = valid & ~invalid_parent[parent_rel_idx]
            #jax.debug.print('valid {v}, parent valid {p}', v=valid, p=~invalid_parent[parent_rel_idx])
            new_node = jax.lax.select(valid, jnp.concatenate([test_pos, jnp.array([parent_idx, cost, goal_just_reached.astype(jnp.int32)])]), jnp.zeros(tree.shape[1]))
            #jax.debug.print('new node {n}, goal reached {g}', n=new_node, g=goal_just_reached)
            tree = tree.at[rrt_idx].set(new_node)
            
            # rewire
            invalid_child = invalid_parent.at[parent_rel_idx].set(True)
            
            new_cost = cost + tree_dist[parent_poss] + INF*invalid_child
            rewire = (new_cost < tree[parent_poss, 3]) & valid 
            
            new_nodes = jnp.where(rewire[:, None], jnp.concatenate([tree[parent_poss, :2], jnp.full((rrt_star_neighbours, 1), rrt_idx), new_cost[:, None], tree[parent_poss, 4][:, None]], axis=1), tree[parent_poss])
            #jax.debug.print('new nodes {n}', n=new_nodes)
            tree = tree.at[parent_poss].set(new_nodes)
            
            new_goal_node = goal_just_reached & valid

            rrt_idx += 1*valid
            goal_reached = goal_reached | new_goal_node
            return (i+1, goal_reached, rrt_idx, tree, key)
            
        val = jax.lax.while_loop(_cond_fun, _body_fun, (0, goal_reached, rrt_idx, tree, key))
        goal_reached, tree = val[1], val[3]    
        
        return tree, goal_reached
            
    def plot_rrt_tree(self, ax, tree, goal_reached=False, goal=None, rrt_reward=None, name="env"):
        
        for n in range(tree.shape[0]):
            if tree[n, 0] == 0.0: break
            ax.scatter(tree[n, 0], tree[n, 1], c='gray')
            parent_idx = tree[n, 2]
            if parent_idx == -1: continue
            parent_idx = int(parent_idx)
            ax.plot([tree[n, 0], tree[parent_idx, 0]], [tree[n, 1], tree[parent_idx, 1]], c='gray', marker='+', alpha=0.75)
    
                
        if goal_reached:
            if rrt_reward is not None: 
                assert goal is not None            
                rewards = []
            
            path_lengths = []
            goal_idx = jnp.argwhere(tree[:,-1]==1)
            
            for g_idx in goal_idx:
                c_idx = g_idx[0]
                path_length = 0.0
                if rrt_reward is not None: rew = 0.0
                while c_idx != 0:
                    #print('cidx', c_idx, 'tree row', tree[c_idx])
                    c_pos = tree[c_idx, :2]
                    p_idx = int(tree[c_idx, 2])
                    p_pos = tree[p_idx, :2]
                    path_length += jnp.linalg.norm(c_pos - p_pos)
                    if rrt_reward is not None: rew += rrt_reward(c_pos, p_pos, goal)
                    ax.plot([c_pos[0], p_pos[0]], [c_pos[1], p_pos[1]], c='r', alpha=0.25)
                    c_idx = p_idx
                path_lengths.append(path_length)
                #print('path_length:', path_length)
                if rrt_reward is not None: 
                    #print('reward:', rew)
                    rewards.append(rew)
        
            if rrt_reward is not None:
                max_rew_idx = jnp.argmax(jnp.array(rewards))
                print(name, ' max reward:', rewards[max_rew_idx], 'path length:', path_lengths[max_rew_idx])  
            else:
                print(name, ' min path length:', jnp.min(jnp.array(path_lengths)))
    
    def plot_map(self,
                 ax: axes.Axes,
                 map_data: jnp.ndarray,) -> None:
        raise NotImplementedError
            
    def plot_agents(
                self,
                ax: axes.Axes,
                pos: jnp.ndarray,
                theta: jnp.ndarray,
                goal: jnp.ndarray,
                done: jnp.ndarray,
                plot_line_to_goal=True,
                colour_agents_by_idx=False,
        ) -> None:
        raise NotImplementedError
    
    def plot_agent_path(
        self,
        ax: axes.Axes,
        x_seq: jnp.ndarray,
        y_seq: jnp.ndarray,
    ) -> None:
        raise NotImplementedError
     
            
    
        
    



import jax
import jax.numpy as jnp
import numpy as np
from .baselines import JaxMARLWrapper, CTRolloutManager
from jaxmarl.environments.multi_agent_env import MultiAgentEnv
from functools import partial

class TransformersCTRolloutManager(CTRolloutManager):
    # wraps envs the rollout manager for managing transformers observations and states matrices 
    
    def __init__(self, env: MultiAgentEnv, batch_size:int):

        # SMAX
        if 'smax' in env.name.lower():
            # For SMAX the simplest approach is to post-process the obs, since they already contain a matrix structure 
            super().__init__(env, batch_size, training_agents=None, preprocess_obs=True)
            self._preprocess_obs = self.smax_obs_vec_to_matrix
            self.global_state = self.smax_global_state

        # SPREAD  
        elif 'spread' in env.name.lower():
            # For SPREAD it is simpler to build the obs from scratch as matrices
            super().__init__(env, batch_size, training_agents=None, preprocess_obs=False)
            if any(issubclass(parent, JaxMARLWrapper) for parent in type(env).__mro__): #apply the wrapped get_obs to the child env class
                self._env._env.get_obs = self.spread_wrapped_get_obs
            else:
                self._env.get_obs = self.spread_wrapped_get_obs
            self.global_state = lambda obs, state: obs['world_state']
        
        else:
            raise NotImplementedError('This implemention currently supports only MPE_spread and SMAX')
        

    @partial(jax.jit, static_argnums=0)
    def smax_obs_vec_to_matrix(self, obs, extra_feats):
        # extract the features relative to others and self and make a matrix
        others_feats = obs[:(self._env.obs_size-len(self._env.own_features))].reshape(-1, len(self._env.unit_features))
        self_feats = obs[-len(self._env.own_features):]
        pad_width = [(0, max(0, others_feats.shape[-1] - self_feats.shape[-1]))]
        self_feats = jnp.pad(self_feats, pad_width, mode='constant', constant_values=0)
        rel_feats = jnp.concatenate((others_feats, self_feats[np.newaxis, :]), axis=0)
        # the last obs vector refers to self
        is_self_feat = jnp.zeros(self.num_allies+self.num_enemies).at[-1].set(1)
        # first are teamates, then enemies and finally the self
        is_agent_feat = jnp.concatenate((jnp.ones(self._env.num_allies-1), jnp.zeros(self._env.num_enemies), jnp.ones(1)))
        feats = jnp.concatenate((
            rel_feats,
            is_agent_feat[:, None],
            is_self_feat[:, None],
        ), axis=1)
        return feats
            
    @partial(jax.jit, static_argnums=0)
    def smax_global_state(self, obs, state):
        # extract the main feats that are defined for all the entities
        main_feats = obs['world_state'][:len(self._env.own_features)*(self._env.num_allies+self._env.num_enemies)]
        main_feats = main_feats.reshape(self._env.num_allies+self._env.num_enemies, -1)
        
        # the other feats are is_agent and unit_type, defined for each entity
        other_feats = obs['world_state'][len(self._env.own_features)*(self._env.num_allies+self._env.num_enemies):]
        other_feats = jnp.swapaxes(other_feats.reshape(-1, self._env.num_allies+self._env.num_enemies), 0,1)
        
        return jnp.concatenate((main_feats, other_feats), axis=1)
    
    @partial(jax.jit, static_argnums=0)
    def spread_wrapped_get_obs(self, state):
        """
        - Obs feats: [d_x, d_y, is_self, is_agent]
        - State feats: [x, y, vel_x, vel_y, is_agent]
        """
        # relative position between agents and other entities
        rel_pos = state.p_pos - state.p_pos[:self._env.num_agents, None, :]
        is_self_feat  = (jnp.arange(self._env.num_entities) == jnp.arange(self._env.num_agents)[:, np.newaxis])
        is_agent_feat = jnp.tile(
            jnp.concatenate((jnp.ones(self._env.num_agents), jnp.zeros(self._env.num_landmarks))),
            (self._env.num_agents, 1)
        )
        feats = jnp.concatenate((
            rel_pos,
            is_self_feat[:, :, None],
            is_agent_feat[:, :, None],
        ), axis=2)

        obs = {
            a:feats[i]
            for i, a in enumerate(self._env.agents)
        }
        
        obs['world_state'] = jnp.concatenate((
            state.p_pos,
            state.p_vel,
            is_agent_feat[0][:, None]
        ), axis=1)
        
        return obs


from typing import Sequence, NamedTuple, Any
from flax.linen.initializers import constant, orthogonal
import jax
import jax.numpy as jnp
import flax.linen as nn
from flax.training.train_state import TrainState
import numpy as np
import optax
import distrax
from jaxmarl import make
from jaxmarl.environments.smax import map_name_to_scenario, Scenario
import time


def batchify(x: dict, agent_list, num_actors):
    x = jnp.stack([x[a] for a in agent_list])
    return x.reshape((num_actors, -1))


def unbatchify(x: dict, agent_list, num_envs, num_actors):
    x = x.reshape((num_actors, num_envs, -1))
    return {a: x[i] for i, a in enumerate(agent_list)}

EXTRA_SCENARIOS = {
    "5m": Scenario(jnp.zeros((10,), dtype=jnp.uint8), num_allies=5, num_enemies=5, smacv2_position_generation=False, smacv2_unit_type_generation=False),
    "50m": Scenario(jnp.zeros((100,), dtype=jnp.uint8), num_allies=50, num_enemies=50, smacv2_position_generation=False, smacv2_unit_type_generation=False),
    "500m": Scenario(jnp.zeros((1000,), dtype=jnp.uint8), num_allies=500, num_enemies=500, smacv2_position_generation=False, smacv2_unit_type_generation=False),
    "5000m": Scenario(jnp.zeros((10000,), dtype=jnp.uint8), num_allies=5000, num_enemies=5000, smacv2_position_generation=False, smacv2_unit_type_generation=False)
}


class ActorCritic(nn.Module):
    action_dim: Sequence[int]
    activation: str = "tanh"

    @nn.compact
    def __call__(self, x):
        if self.activation == "relu":
            activation = nn.relu
        else:
            activation = nn.tanh
        actor_mean = nn.Dense(
            64, kernel_init=orthogonal(np.sqrt(2)), bias_init=constant(0.0)
        )(x)
        actor_mean = activation(actor_mean)
        actor_mean = nn.Dense(
            64, kernel_init=orthogonal(np.sqrt(2)), bias_init=constant(0.0)
        )(actor_mean)
        actor_mean = activation(actor_mean)
        actor_mean = nn.Dense(
            self.action_dim, kernel_init=orthogonal(0.01), bias_init=constant(0.0)
        )(actor_mean)
        pi = distrax.Categorical(logits=actor_mean)

        critic = nn.Dense(
            64, kernel_init=orthogonal(np.sqrt(2)), bias_init=constant(0.0)
        )(x)
        critic = activation(critic)
        critic = nn.Dense(
            64, kernel_init=orthogonal(np.sqrt(2)), bias_init=constant(0.0)
        )(critic)
        critic = activation(critic)
        critic = nn.Dense(1, kernel_init=orthogonal(1.0), bias_init=constant(0.0))(
            critic
        )

        return pi, jnp.squeeze(critic, axis=-1)


def make_benchmark(config):
    if config["MAP_NAME"] in EXTRA_SCENARIOS:
        scenario = EXTRA_SCENARIOS[config["MAP_NAME"]]
    else:
        scenario = map_name_to_scenario(config["MAP_NAME"])
    env = make(config["ENV_NAME"], scenario=scenario, **config["ENV_KWARGS"])
    config["NUM_ACTORS"] = env.num_agents * config["NUM_ENVS"]
    network = ActorCritic(
        env.action_space(env.agents[0]).n, activation=config["ACTIVATION"]
    )

    def benchmark(rng):
        def init_runner_state(rng):
            # INIT NETWORK
            network = ActorCritic(
                env.action_space(env.agents[0]).n, activation=config["ACTIVATION"]
            )
            rng, _rng = jax.random.split(rng)
            init_x = jnp.zeros(env.observation_space(env.agents[0]).shape)
            params = network.init(_rng, init_x)
            # INIT ENV
            rng, _rng = jax.random.split(rng)
            reset_rng = jax.random.split(_rng, config["NUM_ENVS"])
            obsv, env_state = jax.vmap(env.reset)(reset_rng)

            return (params, env_state, obsv, rng)

        def env_step(runner_state, unused):
            params, env_state, last_obs, rng = runner_state

            # SELECT ACTION
            rng, _rng = jax.random.split(rng)

            obs_batch = batchify(last_obs, env.agents, config["NUM_ACTORS"])

            if config["ACTION_SELECTION"] == "nn":
                pi, _ = network.apply(params, obs_batch)
                # transform using the available actions
                avail_actions = jax.vmap(env.get_avail_actions)(env_state)
                avail_actions = jax.lax.stop_gradient(
                    batchify(avail_actions, env.agents, config["NUM_ACTORS"])
                )
                probs = jnp.where(avail_actions, pi.probs, 1e-10)
                probs = probs / probs.sum(axis=-1)[:, None]
                pi = distrax.Categorical(probs=probs)

                action = pi.sample(seed=_rng)
            elif config["ACTION_SELECTION"] == "random":
                avail_actions = jax.vmap(env.get_avail_actions)(env_state)
                avail_actions = jax.lax.stop_gradient(
                    batchify(avail_actions, env.agents, config["NUM_ACTORS"])
                )
                avail_actions = avail_actions.reshape(
                    config["NUM_ENVS"], env.num_agents, -1
                )
                logits = jnp.log(
                    jnp.ones((avail_actions.shape[-1],))
                    / jnp.sum(jnp.ones((avail_actions.shape[-1],)))
                )
                action = jax.random.categorical(
                    _rng, logits=logits, shape=avail_actions.shape[:-1]
                )
                action = action.reshape(config["NUM_ACTORS"], -1)
            else:
                raise ValueError("ACTION_SELECTION must be 'random' or 'nn'")

            env_act = unbatchify(action, env.agents, config["NUM_ENVS"], env.num_agents)
            env_act = {k: v.squeeze() for k, v in env_act.items()}

            # STEP ENV
            rng, _rng = jax.random.split(rng)
            rng_step = jax.random.split(_rng, config["NUM_ENVS"])
            obsv, env_state, _, _, info = jax.vmap(env.step)(
                rng_step, env_state, env_act
            )
            info = jax.tree.map(lambda x: x.reshape((config["NUM_ACTORS"])), info)
            runner_state = (params, env_state, obsv, rng)
            return runner_state, None

        rng, init_rng = jax.random.split(rng)
        runner_state = init_runner_state(init_rng)
        runner_state = jax.lax.scan(env_step, runner_state, None, config["NUM_STEPS"])
        return runner_state

    return benchmark


def main():
    config = {
        "NUM_STEPS": 128,
        "NUM_ENVS": 4,
        "ACTIVATION": "relu",
        "MAP_NAME": "50m",
        "ENV_KWARGS": {
            "map_width": 32,
            "map_height": 32,
        },
        "ENV_NAME": "HeuristicEnemySMAX",
        "NUM_SEEDS": 1,
        "SEED": 0,
        "ACTION_SELECTION": "random"
    }
    benchmark_fn = make_benchmark(config)
    rng = jax.random.PRNGKey(config["SEED"])
    benchmark_jit = jax.jit(benchmark_fn).lower(rng).compile()
    before = time.perf_counter_ns()
    runner_state = jax.block_until_ready(benchmark_jit(rng))
    after = time.perf_counter_ns()
    num_steps = config["NUM_ENVS"] * config["NUM_STEPS"]
    total_time = (after - before) / 1e9
    print(f"Total Time (s): {total_time}")
    print(f"Total Steps: {num_steps}")
    print(f"SPS: {num_steps / total_time}")


if __name__ == "__main__":
    main()



from functools import partial
import jax
import jax.numpy as jnp
from typing import List, Tuple
import itertools
from .common import ALL_DIRECTIONS, Position, Direction


def tree_select(predicate, a, b):
    return jax.tree_util.tree_map(lambda x, y: jax.lax.select(predicate, x, y), a, b)


def compute_view_box(x, y, agent_view_size, height, width):
    """Compute the view box for an agent centered at (x, y)"""
    x_low = x - agent_view_size
    x_high = x + agent_view_size + 1
    y_low = y - agent_view_size
    y_high = y + agent_view_size + 1

    x_low = jax.lax.clamp(0, x_low, width)
    x_high = jax.lax.clamp(0, x_high, width)
    y_low = jax.lax.clamp(0, y_low, height)
    y_high = jax.lax.clamp(0, y_high, height)

    return x_low, x_high, y_low, y_high


def compute_enclosed_spaces(empty_mask: jnp.ndarray) -> jnp.ndarray:
    """
    Compute the enclosed spaces in the environment.
    Each enclosed space is assigned a unique id.
    """
    height, width = empty_mask.shape
    id_grid = jnp.arange(empty_mask.size, dtype=jnp.int32).reshape(empty_mask.shape)
    id_grid = jnp.where(empty_mask, id_grid, -1)

    def _body_fun(val):
        _, curr = val

        def _next_val(pos):
            neighbors = jax.vmap(pos.move_in_bounds, in_axes=(0, None, None))(
                jnp.array(list(Direction)), width, height
            )
            neighbour_values = curr[neighbors.y, neighbors.x]
            self_value = curr[pos.y, pos.x]
            values = jnp.concatenate(
                [neighbour_values, self_value[jnp.newaxis]], axis=0
            )
            new_val = jnp.max(values)
            return jax.lax.select(self_value == -1, self_value, new_val)

        pos_y, pos_x = jnp.meshgrid(
            jnp.arange(height), jnp.arange(width), indexing="ij"
        )

        next_vals = jax.vmap(jax.vmap(_next_val))(Position(x=pos_x, y=pos_y))
        stop = jnp.all(curr == next_vals)
        return stop, next_vals

    def _cond_fun(val):
        return ~val[0]

    initial_val = (False, id_grid)
    _, res = jax.lax.while_loop(_cond_fun, _body_fun, initial_val)
    return res


def mark_adjacent_cells(mask):
    # Shift the mask in four directions: up, down, left, right
    up = jnp.roll(mask, shift=-1, axis=0)
    down = jnp.roll(mask, shift=1, axis=0)
    left = jnp.roll(mask, shift=-1, axis=1)
    right = jnp.roll(mask, shift=1, axis=1)

    # Prevent wrapping by zeroing out the rolled values at the boundaries
    up = up.at[-1, :].set(False)
    down = down.at[0, :].set(False)
    left = left.at[:, -1].set(False)
    right = right.at[:, 0].set(False)

    # Combine the original mask with the shifted versions
    expanded_mask = mask | up | down | left | right

    return expanded_mask


def get_closest_true_pos_no_directions(arr: jnp.ndarray, pos: Position) -> Position:
    height, width = arr.shape

    y, x = pos.y, pos.x
    yy, xx = jnp.meshgrid(jnp.arange(height), jnp.arange(width), indexing="ij")

    dist = jnp.abs(yy - y) + jnp.abs(xx - x)
    dist = jnp.where(arr, dist, jnp.inf)

    min_idx = jnp.argmin(dist)
    min_y, min_x = jnp.divmod(min_idx, width)

    is_valid = jnp.any(arr)

    return Position(x=min_x, y=min_y), is_valid


class OvercookedPathPlanner:
    def __init__(self, move_area: jnp.ndarray):
        self._precompute(move_area)

    @partial(jax.jit, static_argnums=(0,))
    def get_closest_target_pos(
        self, targets: jnp.ndarray, pos: Position, dir: Direction
    ) -> Tuple[Position, bool]:
        pos_lookup_idx = self.position_to_idx[pos.y, pos.x]

        def _compute_min_moves():
            min_moves = self.precomputed_min_moves[pos_lookup_idx, dir]
            return self._get_pos_from_min_moves_grid(min_moves, targets)

        is_target = targets[pos.y, pos.x]
        is_allowed_pos = pos_lookup_idx != -1
        return jax.lax.cond(
            is_target | ~is_allowed_pos,
            lambda: (pos, is_allowed_pos),
            _compute_min_moves,
        )

    def _precompute(self, move_area: jnp.ndarray):
        pos_idx = jnp.argwhere(move_area)
        num_pos = pos_idx.shape[0]
        positions = Position(y=pos_idx[:, 0], x=pos_idx[:, 1])

        self.precomputed_min_moves = jax.vmap(
            jax.vmap(self._compute_min_moves, in_axes=(None, 0, None)),
            in_axes=(0, None, None),
        )(positions, ALL_DIRECTIONS, move_area)

        self.position_to_idx = (
            jnp.full_like(move_area, -1, dtype=jnp.int32)
            .at[positions.y, positions.x]
            .set(jnp.arange(num_pos))
        )

        # print("Precomputation done")
        # print("Number of positions:", num_pos)
        # print("Shape of precomputed_min_moves:", self.precomputed_min_moves.shape)
        # print("Shape of position_to_idx:", self.position_to_idx.shape)

    @classmethod
    @partial(jax.jit, static_argnums=(0,))
    def get_closest_target_pos_static(
        cls, move_area: jnp.ndarray, targets: jnp.ndarray, pos: Position, dir: Direction
    ) -> Tuple[Position, bool]:

        def _compute_min_moves(pos, dir):
            min_moves = cls._compute_min_moves(pos, dir, move_area)
            return cls._get_pos_from_min_moves_grid(min_moves, targets)

        return jax.lax.cond(
            targets[pos.y, pos.x],
            lambda p, _d: (p, True),
            _compute_min_moves,
            pos,
            dir,
        )

    @staticmethod
    def _get_pos_from_min_moves_grid(
        min_moves: jnp.ndarray, targets: jnp.ndarray
    ) -> Tuple[Position, bool]:
        min_moves_targets = jnp.where(targets, min_moves, jnp.inf)

        min_idx = jnp.argmin(min_moves_targets)
        min_y, min_x = jnp.divmod(min_idx, min_moves.shape[1])

        is_valid = jnp.any(jnp.isfinite(min_moves_targets))

        return Position(x=min_x, y=min_y), is_valid

    @staticmethod
    @jax.jit
    def _compute_min_moves(
        pos: Position, dir: Direction, mask: jnp.ndarray
    ) -> jnp.ndarray:
        assert mask.ndim == 2 and mask.dtype == jnp.bool_
        H, W = mask.shape

        ys, xs = jnp.meshgrid(jnp.arange(H), jnp.arange(W), indexing="ij")

        def _obstacle_ahead(pos, dir):
            new_pos, is_valid = pos.checked_move(dir, W, H)
            obstacle_ahead = ~mask[new_pos.y, new_pos.x]
            return ~is_valid | obstacle_ahead

        obstacle_ahead = jax.vmap(_obstacle_ahead, in_axes=(None, 0), out_axes=-1)(
            Position(x=xs, y=ys), ALL_DIRECTIONS
        )
        obstacle_ahead &= mask[..., jnp.newaxis]

        def cond_fun(loop_carry):
            _, changed = loop_carry
            return changed

        def body_fun(loop_carry):
            dist, _ = loop_carry
            changed = False

            min_across_last_dim = jnp.min(dist, axis=-1)

            def _move_dir(dir):
                moved_dist = jnp.full_like(min_across_last_dim, jnp.inf)
                if dir == Direction.UP:
                    moved_dist = moved_dist.at[:-1, :].set(min_across_last_dim[1:, :])
                elif dir == Direction.DOWN:
                    moved_dist = moved_dist.at[1:, :].set(min_across_last_dim[:-1, :])
                elif dir == Direction.RIGHT:
                    moved_dist = moved_dist.at[:, 1:].set(min_across_last_dim[:, :-1])
                elif dir == Direction.LEFT:
                    moved_dist = moved_dist.at[:, :-1].set(min_across_last_dim[:, 1:])
                return jnp.where(mask, moved_dist, jnp.inf)

            dist_up = _move_dir(Direction.UP)
            dist_down = _move_dir(Direction.DOWN)
            dist_right = _move_dir(Direction.RIGHT)
            dist_left = _move_dir(Direction.LEFT)

            dist_new = jnp.stack([dist_up, dist_down, dist_right, dist_left], axis=-1)

            blocked_new_dist = jnp.where(
                obstacle_ahead, min_across_last_dim[..., jnp.newaxis], jnp.inf
            )

            dist_new = jnp.minimum(dist_new, blocked_new_dist)
            dist_new += 1
            dist_updated = jnp.minimum(dist, dist_new)

            changed = jnp.any(dist_updated != dist)

            return dist_updated, changed

        initial_dist = jnp.full((H, W, 4), jnp.inf, dtype=jnp.float32)
        initial_dist = initial_dist.at[pos.y, pos.x, dir].set(0)

        dist_final, _ = jax.lax.while_loop(cond_fun, body_fun, (initial_dist, True))

        def _compute_min_cost(pos, dir):
            new_pos, is_valid = pos.checked_move(dir, W, H)
            opposite_dir = Direction.opposite(dir)

            return jnp.where(
                is_valid,
                dist_final[new_pos.y, new_pos.x, opposite_dir],
                jnp.inf,
            )

        min_cost_to_target = jax.vmap(
            _compute_min_cost, in_axes=(None, 0), out_axes=-1
        )(Position(x=xs, y=ys), ALL_DIRECTIONS)
        min_cost_to_target = jnp.min(min_cost_to_target, axis=-1)

        # we only care about target cells
        min_cost_to_target = jnp.where(mask, jnp.inf, min_cost_to_target)

        return min_cost_to_target



"""
This class models the game dynamics of Hanabi (reset and step of the game).
"""

import numpy as np
import jax
import jax.numpy as jnp
from jax import lax
import chex
from flax import struct
from typing import Tuple, Dict
from functools import partial
from jaxmarl.environments.multi_agent_env import MultiAgentEnv


@struct.dataclass
class State:
    deck: chex.Array
    discard_pile: chex.Array
    fireworks: chex.Array
    player_hands: chex.Array
    info_tokens: chex.Array
    terminal: bool
    life_tokens: chex.Array
    card_knowledge: chex.Array
    colors_revealed: chex.Array
    ranks_revealed: chex.Array
    num_cards_dealt: int
    num_cards_discarded: int
    cur_player_idx: chex.Array
    out_of_lives: bool
    last_round_count: int
    bombed: bool
    remaining_deck_size: chex.Array
    turn: int
    score: int


class HanabiGame(MultiAgentEnv):

    def __init__(
        self,
        num_agents=2,
        num_colors=5,
        num_ranks=5,
        hand_size=5,
        max_info_tokens=8,
        max_life_tokens=3,
        num_cards_of_rank=np.array([3, 2, 2, 2, 1]),
        color_map=["R", "Y", "G", "W", "B"],
    ):
        super().__init__(num_agents)

        self.num_agents = num_agents
        self.agent_range = jnp.arange(num_agents)
        self.num_colors = num_colors
        self.num_ranks = num_ranks
        self.hand_size = hand_size
        self.max_info_tokens = max_info_tokens
        self.max_life_tokens = max_life_tokens
        self.num_cards_of_rank = num_cards_of_rank
        self.deck_size = np.sum(num_cards_of_rank) * num_colors
        self.color_map = color_map

        # action ranges - useful to know
        self.discard_action_range = jnp.arange(
            0,
            self.hand_size
        )
        self.play_action_range = jnp.arange(
            self.hand_size,
            2 * self.hand_size
        )
        self.color_action_range = jnp.arange(
            2 * self.hand_size,
            2 * self.hand_size + (self.num_agents - 1) * self.num_colors
        )
        self.rank_action_range = jnp.arange(
            2 * self.hand_size + (self.num_agents - 1) * self.num_colors,
            2 * self.hand_size + (self.num_agents - 1) * (self.num_colors + self.num_ranks)
        )

    @partial(jax.jit, static_argnums=[0])
    def get_first_state(self, deck: chex.Array) -> State:
        """Get the initial state of the game"""

        def _deal_cards(aidx, unused):
            """Deals cards to players from top of deck"""
            # top of deck is first array element
            start = aidx * self.hand_size
            hand = lax.dynamic_slice(
                deck, (start, 0, 0), (self.hand_size, self.num_colors, self.num_ranks)
            )

            return aidx + 1, hand

        _, hands = lax.scan(_deal_cards, 0, None, self.num_agents)
        num_cards_dealt = self.num_agents * self.hand_size

        # start off with all (color, rank) combinations being possible for all cards
        card_knowledge = jnp.ones(
            (self.num_agents, self.hand_size, (self.num_colors * self.num_ranks))
        )
        colors_revealed = jnp.zeros((self.num_agents, self.hand_size, self.num_colors))
        ranks_revealed = jnp.zeros((self.num_agents, self.hand_size, self.num_ranks))

        # init discard pile
        discard_pile = jnp.zeros_like(deck)
        num_cards_discarded = 0

        # remove dealt cards from deck
        deck = deck.at[:num_cards_dealt].set(
            jnp.zeros((self.num_colors, self.num_ranks))
        )
        remaining_deck_size = jnp.zeros(self.deck_size).at[:-num_cards_dealt].set(1)

        # thermometer encoded
        life_tokens = jnp.ones(self.max_life_tokens)
        info_tokens = jnp.ones(self.max_info_tokens).astype(int)
        fireworks = jnp.zeros((self.num_colors, self.num_ranks))

        # other state variable inits
        cur_player_idx = jnp.zeros(self.num_agents).at[0].set(1)
        terminal = False
        out_of_lives = False
        bombed = False
        last_round_count = 0

        state = State(
            deck=deck,
            discard_pile=discard_pile,
            fireworks=fireworks,
            player_hands=hands,
            info_tokens=info_tokens,
            terminal=terminal,
            life_tokens=life_tokens,
            card_knowledge=card_knowledge,
            colors_revealed=colors_revealed,
            ranks_revealed=ranks_revealed,
            num_cards_dealt=num_cards_dealt,
            num_cards_discarded=num_cards_discarded,
            cur_player_idx=cur_player_idx,
            out_of_lives=out_of_lives,
            last_round_count=last_round_count,
            bombed=bombed,
            remaining_deck_size=remaining_deck_size,
            turn=0,
            score=0,
        )

        return state

    @partial(jax.jit, static_argnums=[0])
    def reset_game(self, key: chex.PRNGKey) -> State:
        """Create a random deck and return the first state of the game"""

        # get all possible (colour, rank) pairs, including repetitions given num_cards_of_rank
        colors = jnp.arange(self.num_colors)
        ranks = jnp.arange(self.num_ranks)
        ranks = jnp.repeat(ranks, self.num_cards_of_rank)
        color_rank_pairs = jnp.dstack(jnp.meshgrid(colors, ranks)).reshape(-1, 2)
        # randomly shuffle (colour, rank) pairs
        key, _key = jax.random.split(key)
        shuffled_pairs = jax.random.permutation(_key, color_rank_pairs, axis=0)
        # generate one-hot encoded deck
        deck = self._one_hot_encode_deck(shuffled_pairs)
        return self.get_first_state(deck)

    @partial(jax.jit, static_argnums=[0])
    def reset_game_from_deck(self, deck: chex.Array) -> State:
        """Reset the game from a given deck of one-hot encoded cards."""
        return self.get_first_state(deck)

    @partial(jax.jit, static_argnums=[0])
    def reset_game_from_deck_of_pairs(self, deck_of_pairs: chex.Array):
        """Reset the game from a given deck of (color, value) pairs."""
        deck = self._one_hot_encode_deck(deck_of_pairs)
        return self.get_first_state(deck)

    @partial(jax.jit, static_argnums=[0])
    def step_game(
        self,
        state: State,
        aidx: int,
        action: int,
    ) -> Tuple[State, int]:
        """
        Execute the current player's action and its consequences
        """
        # check move type
        is_discard = self._is_discard(action)
        is_hint = self._is_hint(action)
        # initialise reward for move
        reward = 0

        def _discard_play_fn(state: State, action: int):
            """Discard or play selected card according to action selection"""
            # get hand and card info
            hand_before = state.player_hands.at[aidx].get()
            card_idx = (
                (is_discard * action)
                + (jnp.logical_not(is_discard) * (action - self.hand_size))
            ).astype(int)
            card = hand_before.at[card_idx].get()

            # gain an info token for discarding if discard action
            infos_remaining = jnp.sum(state.info_tokens)
            infos_depleted = infos_remaining < self.max_info_tokens
            new_infos = infos_remaining + (is_discard * infos_depleted)
            info_tokens = jnp.where(
                new_infos > 0,
                state.info_tokens.at[new_infos - 1].set(1),
                state.info_tokens
            )

            # play selected card if play action
            color, rank = jnp.nonzero(card, size=1)
            color_fireworks = state.fireworks.at[color].get()
            is_valid_play = rank == jnp.sum(color_fireworks)
            make_play = jnp.logical_and(
                is_valid_play, jnp.logical_not(is_discard)
            ).squeeze(0)

            # gain another info token if completed a color
            is_final_card = jnp.logical_and(
                make_play, rank == self.num_ranks - 1
            ).squeeze(0)
            infos_remaining = jnp.sum(info_tokens)
            infos_depleted = infos_remaining < self.max_info_tokens
            new_infos = infos_remaining + (is_final_card * infos_depleted)
            info_tokens = jnp.where(
                new_infos > 0,
                info_tokens.at[new_infos - 1].set(1),
                info_tokens
            )

            # increment fireworks if valid play action
            color_fireworks = color_fireworks.at[
                0, jnp.sum(color_fireworks).astype(int)
            ].set(make_play)
            fireworks = state.fireworks.at[color].set(color_fireworks)

            # the card must be discarded if action is discard or the play action is not valid
            discard_card = ((~is_valid_play) | (is_discard)).squeeze(0)
            discarded_card = jnp.zeros_like(card) + (discard_card * card)
            discard_pile = state.discard_pile.at[state.num_cards_discarded].set(
                discarded_card
            )
            num_cards_discarded = state.num_cards_discarded + discard_card

            # remove life token if invalid play
            life_lost = jnp.logical_not(
                jnp.logical_or(is_valid_play, is_discard)
            ).squeeze(0)
            num_life_tokens = jnp.sum(state.life_tokens).astype(int)
            life_tokens = jnp.where(
                life_lost,
                state.life_tokens.at[num_life_tokens - 1].set(0),
                state.life_tokens
            )

            # color hint knowledge removal
            player_colors_revealed = state.colors_revealed.at[aidx].get()
            player_colors_revealed = jnp.delete(
                player_colors_revealed, card_idx, axis=0, assume_unique_indices=True
            )
            player_colors_revealed = jnp.append(
                player_colors_revealed, jnp.zeros((1, self.num_colors)), axis=0
            )
            colors_revealed = state.colors_revealed.at[aidx].set(player_colors_revealed)
            # rank hint knowledge removal
            player_ranks_revealed = state.ranks_revealed.at[aidx].get()
            player_ranks_revealed = jnp.delete(
                player_ranks_revealed, card_idx, axis=0, assume_unique_indices=True
            )
            player_ranks_revealed = jnp.append(
                player_ranks_revealed, jnp.zeros((1, self.num_ranks)), axis=0
            )
            ranks_revealed = state.ranks_revealed.at[aidx].set(player_ranks_revealed)

            # deal a new card
            # check if in last round
            in_last_round = state.last_round_count > 0
            # deal empty card from top of deck if in last round
            new_card = state.deck.at[state.num_cards_dealt].get()
            hand_without_old = jnp.delete(
                hand_before, card_idx, axis=0, assume_unique_indices=True
            )
            new_hand = jnp.append(hand_without_old, new_card[jnp.newaxis, :, :], axis=0)
            hands = state.player_hands.at[aidx].set(new_hand)
            deck = state.deck.at[state.num_cards_dealt].set(jnp.zeros_like(card))
            # don't increment if in last round
            num_cards_dealt = lax.select(
                in_last_round, state.num_cards_dealt, state.num_cards_dealt + 1
            )
            remaining_deck_size = state.remaining_deck_size.at[-num_cards_dealt].set(0)

            # remove knowledge of selected card
            player_knowledge = state.card_knowledge.at[aidx].get()
            player_knowledge = jnp.delete(
                player_knowledge, card_idx, axis=0, assume_unique_indices=True
            )
            new_card_knowledge = jnp.where(
                new_card.any(),
                jnp.ones((1, self.num_colors * self.num_ranks)),
                jnp.zeros((1, self.num_colors * self.num_ranks)),
            )
            player_knowledge = jnp.append(
                player_knowledge,
                new_card_knowledge,
                axis=0,
            )
            card_knowledge = state.card_knowledge.at[aidx].set(player_knowledge)

            return state.replace(
                deck=deck,
                discard_pile=discard_pile,
                player_hands=hands,
                card_knowledge=card_knowledge,
                colors_revealed=colors_revealed,
                ranks_revealed=ranks_revealed,
                fireworks=fireworks,
                info_tokens=info_tokens,
                life_tokens=life_tokens,
                num_cards_dealt=num_cards_dealt,
                num_cards_discarded=num_cards_discarded,
                remaining_deck_size=remaining_deck_size,
            )

        def _hint_fn(state: State, action: int):
            is_color_hint = self._is_hint_color(action)
            is_rank_hint = self._is_hint_rank(action)

            hint_player, hint_idx = self._get_target_player_and_hint_index(aidx, action)

            hint_color = jnp.zeros(self.num_colors, dtype=int).at[hint_idx].set(1) * is_color_hint
            hint_rank = jnp.zeros(self.num_ranks, dtype=int).at[hint_idx].set(1) * is_rank_hint

            # get current card knowledge of relevant player
            cur_knowledge = state.card_knowledge.at[hint_player].get()

            # get relevant player's hand
            cards = state.player_hands.at[hint_player].get()
            card_colors = jnp.sum(cards, axis=2)
            card_ranks = jnp.sum(cards, axis=1)

            # check which cards have hinted color/rank
            color_hint_matches = jnp.matmul(card_colors, hint_color)
            rank_hint_matches = jnp.matmul(card_ranks, hint_rank)

            negative_color_hints = jnp.outer(1 - color_hint_matches, hint_color)
            negative_color_hints = jnp.repeat(
                negative_color_hints, self.num_colors, axis=1
            ).reshape(cur_knowledge.shape)
            negative_rank_hints = jnp.outer(1 - rank_hint_matches, hint_rank)
            negative_rank_hints = jnp.repeat(
                negative_rank_hints, self.num_ranks, axis=0
            ).reshape(cur_knowledge.shape)

            color_mask = (color_hint_matches * jnp.ones((self.num_colors, self.hand_size))).transpose()
            rank_mask = (rank_hint_matches * jnp.ones((self.num_ranks, self.hand_size))).transpose()

            color_hints = color_mask * (1 - hint_color * jnp.ones((self.hand_size, self.num_colors)))
            color_hints = jnp.repeat(color_hints, self.num_colors, axis=1).reshape(
                cur_knowledge.shape
            )
            rank_hints = rank_mask * (1 - hint_rank * jnp.ones((self.hand_size, self.num_ranks)))
            rank_hints = jnp.repeat(rank_hints, self.num_ranks, axis=0).reshape(
                cur_knowledge.shape
            )

            total_color_hint = color_hints + negative_color_hints
            total_rank_hint = rank_hints + negative_rank_hints

            new_color_knowledge = (
                cur_knowledge - (is_color_hint * total_color_hint)
            ).clip(min=0)
            new_rank_knowledge = (
                cur_knowledge - (is_rank_hint * total_rank_hint)
            ).clip(min=0)
            new_knowledge = jnp.where(
                is_color_hint, new_color_knowledge, new_rank_knowledge
            )
            card_knowledge = state.card_knowledge.at[hint_player].set(new_knowledge)

            colors_revealed_player = jnp.outer(color_hint_matches, hint_color)
            colors_revealed_player = (
                state.colors_revealed.at[hint_player].get() + colors_revealed_player
            ).clip(max=1)
            colors_revealed = state.colors_revealed.at[hint_player].set(
                colors_revealed_player
            )
            ranks_revealed_player = jnp.outer(rank_hint_matches, hint_rank)
            ranks_revealed_player = (
                state.ranks_revealed.at[hint_player].get() + ranks_revealed_player
            ).clip(max=1)
            ranks_revealed = state.ranks_revealed.at[hint_player].set(
                ranks_revealed_player
            )

            # remove an info token
            num_info_tokens = jnp.sum(state.info_tokens).astype(int) - 1
            info_tokens = jnp.arange(self.max_info_tokens)
            info_tokens = (info_tokens < num_info_tokens).astype(int)
            return state.replace(
                card_knowledge=card_knowledge,
                info_tokens=info_tokens,
                colors_revealed=colors_revealed,
                ranks_revealed=ranks_revealed,
            )

        # update fireworks
        fireworks_before = jnp.sum(state.fireworks, axis=(0, 1))
        state = lax.cond(is_hint, _hint_fn, _discard_play_fn, state, action)
        fireworks_after = jnp.sum(state.fireworks, axis=(0, 1))

        # check if lives left
        num_lives = jnp.sum(state.life_tokens)
        out_of_lives = num_lives == 0

        # check if terminal
        game_won = fireworks_after == (self.num_colors * self.num_ranks)
        deck_empty = state.num_cards_dealt >= self.deck_size
        last_round_count = state.last_round_count + deck_empty
        last_round_done = last_round_count == self.num_agents + 1
        terminal = jnp.logical_or(
            jnp.logical_or(state.out_of_lives, game_won), last_round_done
        )

        # define reward as difference in fireworks
        reward += jnp.logical_not(out_of_lives) * (fireworks_after - fireworks_before)
        # bomb-0 scoring
        reward -= out_of_lives * fireworks_after * jnp.logical_not(state.bombed)
        bombed = jnp.logical_or(out_of_lives, state.bombed)
        aidx = (aidx + 1) % self.num_agents

        cur_player_idx = jnp.zeros(self.num_agents).at[aidx].set(1)

        return (
            state.replace(
                terminal=terminal,
                cur_player_idx=cur_player_idx,
                out_of_lives=out_of_lives,
                last_round_count=last_round_count,
                bombed=bombed,
                turn=state.turn + 1,
                score=state.score + reward.astype(int),
            ),
            reward,
        )

    @partial(jax.jit, static_argnums=[0])
    def _one_hot_encode_deck(self, deck_of_pairs: chex.Array) -> chex.Array:
        """Generates one-hot card encodings given (color, rank) pairs."""

        def _gen_cards(aidx, _):
            color, rank = deck_of_pairs[aidx]
            card = jnp.zeros((self.num_colors, self.num_ranks))
            card = card.at[color, rank].set(1)

            return aidx + 1, card

        _, deck = lax.scan(_gen_cards, 0, None, self.deck_size)
        return deck

    @partial(jax.jit, static_argnums=[0])
    def _get_target_player_and_hint_index(self, aidx: int, action: int):
        """
        Determines the target player and the hint index based on the action of the current agent.
        In case you need a one hot encoded representation use:
        jax.nn.one_hot(hint_idx, num_classes=self.)

        Parameters:
        aidx (int): The index of the agent currently taking an action.
        Integer value that uniquely identifies each agent in the game.

        action (int): The raw, unnormalized action in the range 0 to num_actions.
        Represents the specific action that the agent has decided to take.

        Returns:
        target_player (int): The player who is the target of the hint.
        hint_idx (int): The index of the hint that was given.
        For example if the hint is about the value '1', which is the first possible rank action, the hint_idx will be 0.
        Similarly, if color map is ["R", "Y", "G", "W", "B"], these map to values [0, 1, 2, 3, 4] respectively.

        Note:
        This function assumes that the action given is a hint action.
        If an action that is not a hint is given, the behavior of this function will be unexpected.
        """
        is_hint_color = self._is_hint_color(action)
        # adjust the action - we start with a zero index removing the preceeding actions
        action_idx = action - 2 * self.hand_size
        action_idx = jnp.where(
            is_hint_color,
            action_idx,
            action_idx - (self.num_agents - 1) * self.num_colors
        )

        # get the index of hint (rank/color) played
        hint_idx = jnp.where(
            is_hint_color,
            jnp.mod(action_idx, self.num_colors),
            jnp.mod(action_idx, self.num_ranks)
        )

        # get the player to hint.
        target_player_absolute = jnp.where(
            is_hint_color,
            jnp.floor_divide(action_idx, self.num_colors),
            jnp.floor_divide(action_idx, self.num_ranks)
        )
        # adjust for the player who is hinting - wrap around
        target_player = jnp.mod(target_player_absolute + aidx + 1, self.num_agents).astype(int)

        return target_player, hint_idx

    @partial(jax.jit, static_argnums=[0])
    def _is_discard(self, action: int):
        """
        Determines is the action is the discard action where action is the integer value of the action.
        The ranges are defined in `self.discard_action_range`.
        """
        return jnp.logical_and(
            action >= self.discard_action_range[0],
            action <= self.discard_action_range[-1],
        )

    @partial(jax.jit, static_argnums=[0])
    def _is_play(self, action: int):
        """
        Determines is the action is the discard action where action is the integer value of the action.
        The ranges are defined in `self.play_action_range`.
        """
        return jnp.logical_and(
            action >= self.play_action_range[0],
            action <= self.play_action_range[-1],
        )

    @partial(jax.jit, static_argnums=[0])
    def _is_hint_color(self, action: int):
        """
        Determines is the action is the play action where action is the integer value of the action.
        The ranges are defined in `self.color_action_range`.
        """
        return jnp.logical_and(
            action >= self.color_action_range[0],
            action <= self.color_action_range[-1],
        )

    @partial(jax.jit, static_argnums=[0])
    def _is_hint_rank(self, action: int):
        """
        Determines is the action is the play action where action is the integer value of the action.
        The ranges are defined in `self.rank_action_range`.
        """
        return jnp.logical_and(
            action >= self.rank_action_range[0],
            action <= self.rank_action_range[-1],
        )

    @partial(jax.jit, static_argnums=[0])
    def _is_hint(self, action: int):
        """
        Determines is the action is the hint action where action is the integer value of the action.
        """
        return jnp.logical_or(
            self._is_hint_color(action),
            self._is_hint_rank(action),
        )

    @partial(jax.jit, static_argnums=[0])
    def _is_play_or_discard(self, action: int):
        """
        Determines if the action is either a play or discard action. Action is in the integer representation.
        """
        return jnp.logical_or(
            self._is_play(action),
            self._is_discard(action),
        )



"""
Short introduction to the package.

## Abstract base class
Uses the PettingZoo Parallel API. All agents act synchronously, with actions, 
observations, returns and dones passed as dictionaries keyed by agent names. 
The code can be found in `multiagentgymnax/multi_agent_env.py`

The class follows an identical structure to that of `gymnax` with one execption. 
The class is instatiated with `num_agents`, defining the number of agents within the environment.

## Environment loop
Below is an example of a simple environment loop, using random actions.

"""

import jax
import jax.numpy as jnp
import flax.linen as nn
import distrax
from jaxmarl import make
from jaxmarl.environments.smax import map_name_to_scenario
from jaxmarl.environments.smax.heuristic_enemy import (
    create_heuristic_policy,
    get_heuristic_policy_initial_state,
)
from jaxmarl.viz.visualizer import Visualizer, SMAXVisualizer
import os
from typing import Sequence

os.environ["TF_CUDNN_DETERMINISTIC"] = "1"
# Parameters + random keys
max_steps = 15
key = jax.random.PRNGKey(2)
key, key_r, key_a, key_p = jax.random.split(key, 4)


class LearnedPolicy(nn.Module):
    action_dim: Sequence[int]
    activation: str = "tanh"

    @nn.compact
    def __call__(self, x):
        if self.activation == "relu":
            activation = nn.relu
        else:
            activation = nn.tanh
        actor_mean = nn.Dense(self.action_dim)(x)
        actor_mean = activation(actor_mean)
        pi = distrax.Categorical(logits=actor_mean)

        critic = nn.Dense(1)(x)
        critic = activation(critic)

        return pi, jnp.squeeze(critic, axis=-1)


def init_policy(env, rng):
    network = LearnedPolicy(env.action_space(env.agents[0]).n, activation="tanh")
    rng, _rng = jax.random.split(rng)
    init_x = jnp.zeros(env.observation_space(env.agents[0]).shape)
    params = network.init(_rng, init_x)
    return params


# Instantiate environment
with jax.disable_jit(False):
    scenario = map_name_to_scenario("5m_vs_6m")
    env = make(
        "HeuristicEnemySMAX",
        # attack_mode="random",
        scenario=scenario,
        use_self_play_reward=False,
        walls_cause_death=True,
        see_enemy_actions=False,
        # num_allies=3,
        # num_enemies=5,
        # smacv2_position_generation=True,
        # smacv2_unit_type_generation=True,
        action_type="continuous",
        observation_type="conic"
    )
    # env = make("SMAX")
    # params = init_policy(env, key_p)
    # learned_policy = LearnedPolicy(env.action_space(env.agents[0]).n, activation="tanh")
    # env = make("LearnedPolicyEnemySMAX", params=params, policy=learned_policy)
    obs, state = env.reset(key_r)
    print("list of agents in environment", env.agents)

    # Sample random actions
    key_a = jax.random.split(key_a, env.num_agents)
    actions = {
        agent: env.action_space(agent).sample(key_a[i])
        for i, agent in enumerate(env.agents)
    }
    print("example action dict", actions)

    policy_states = {
        agent: get_heuristic_policy_initial_state() for agent in env.agents
    }
    policy = create_heuristic_policy(env, 0, shoot=True, attack_mode="closest")
    enemy_policy = create_heuristic_policy(env, 1, shoot=True, attack_mode="closest")
    state_seq = []
    returns = {a: 0 for a in env.agents}
    for i in range(max_steps):
        # Iterate random keys and sample actions
        key, key_s, key_seq = jax.random.split(key, 3)
        key_a = jax.random.split(key_seq, env.num_agents)
        actions = {}
        # for i, agent in enumerate(env.agents):
        #     p = policy if i < env.num_allies else enemy_policy
        #     action, policy_state = p(key_a[i], policy_states[agent], obs[agent])
        #     policy_states[agent] = policy_state
        #     actions[agent] = action

        # actions = {agent: jnp.array(1) for agent in env.agents}
        actions = {agent: env.action_space(agent).sample(key_a[i]) for i, agent in enumerate(env.agents)}
        state_seq.append((key_s, state, actions))
        # Step environment
        avail_actions = env.get_avail_actions(state)
        obs, state, rewards, dones, infos = env.step(key_s, state, actions)
        print(f"Actions: {actions}")
        returns = {a: returns[a] + rewards[a] for a in env.agents}
        if dones["__all__"]:
            print(f"Returns: {returns}")
print(f"Returns: {returns}")
viz = SMAXVisualizer(env, state_seq)

viz.animate(view=False, save_fname="output.gif")
